/*
 *  Copyright  (c)  L.Allison, C.S.Wallace, C.N.Yee.
 *     Dept. Computer Science, Monash University, Australia 3168,    May 1991.
 *  1. These routines may only be used for non-commercial, non-classified,
 *     research purposes.
 *  2. No liability of any kind is accepted for their use.
 *  3. As a courtesy, please inform us of any interesting uses of the routines,
 *     of any bugs that you find in them and of how to correct
 *     such bugs, if known.
 *  4. Do not delete this copyright notice.
 *  5. Please include the following references in any publication based wholly
 *     or in part on the use of these routines or on the ideas contained in them:
 *   . Allison L. & C.N.Yee.
 *        Minimum Message Length Encoding and the Comparison of Macro-Molecules.
 *        Bull. Math. Bio. 52(3) 431-453 1990
 *   . Allison L., C.S.Wallace & C.N.Yee. Finite-State Models in the Alignment
 *        of Macromolecules. J. Mol. Evol. (1992) 35:77-89
 *  6. You may distribute the routines to others if and only if
 *     they agree to accept and maintain these conditions and no fee is charged.
 *  7. There is no rule 7.
 */
/*	"1State.c"
 *
 *	Compute One-State r-theory:   all weighted alignments
 *	after idea of CSW
 *
 */
#include <stdio.h>
#include <math.h>
#include "macro.h"
#include "Strings.h"
#include "rtheory.h"

#define K	3	/* degree of freedom = no. arcs - 1 */

/*
 *	some global variables
 */
static symbol   *a, *b ;	/* the two sequences */
static int m,n ; /* length of a and b respectively */
static float costMatch, costChange, costIns, costDel ;
AlphabetTable	*Table = &DNASymbol ;

/*
 *	some routines to compute the no. of bits to send the parametes,
 *	length etc.
 */

/*
 *	Gives log2 of gamma(n)
 */
static float LogGamma(n)
	float	n ;
{
	double	sum ;

	sum = 0 ;
	while (n > 2)
		sum += log2((double) --n) ;
	if (n <= 2) 
		if ((n>1.25) && (n<1.75)) sum += log2(sqrt((double)3.141593)/2.0) ;
	return (float) sum ;
		
}

double GammaApprox(n)  /*very rough !*/
	float n ;
/* Gamma(N) = (N-1)!   Gamma continuous fn */
{
	if (n <=2) {
		if ((n>1.25) && (n<1.75))  return sqrt((double)3.141593)/2 ;
		else return (double) 1 ;
	} else /*n>2*/ return (n-1)*GammaApprox(n-1) ;
}

OS_printparameters(P)
	OneStatePBlock *P ;
{ printf("PM=%3.3f PC=%3.3f PID=%3.3f",P->match,P->change,P->indel) ;}

/*
 *	Calculate cost of sending one alignment by an adaptive code,
 *	estimate for fractional chars!  using gamma instead of factorial
 */
static float AdaptiveML(/*frequencies:*/ fm/*atch*/, fc/*hange*/, fid/*indels*/)
	float fm, fc, fid ;
{
	float N ;

	N=fm+fc+fid;
	
	return ((float)( LogGamma(N+K)-LogGamma((float)K)
		- LogGamma(fm+1)
		- LogGamma(fc+1)
		- LogGamma(fid+1)
		+ fm*2 + fc*log2((double)12) + fid*3 /*the chars*/
		+ U(fm + fc + fid)
		)) ;
}

/*---------------------------------------------------------------------------*/
static float sum( C )
	cell *C ;
{	return C->match+C->change+C->indel ; }

float OS_messagelength(C)
	cell *C ;
{ 
	return (float)(paramcost3(C->match, C->change, C->indel) /*to state params*/
		+ C->MesLen     /*the union of all alignments by fixed code*/
		+ U(C->match+C->change+C->indel) );
}

static plot(A,B,D, result, ofile)
	string *A, *B ;
	float D[][MATRIXSIZE] ;
	cell  *result ;
	FILE  *ofile ;
{
	int i, j ;
	float ml, rth, nullth, oddsratio, pcost, lencost, charcost ;
	char ch ;

	printf("Fwd[i,j]+Rev[i+1,j+1]-MML\n");
	rth    = OS_messagelength(result);
	nullth = (float)U((float)m+n)+DiffCode(m,n)+(m+n)*2;
	lencost =(float)U( sum(result) );
	charcost=result->match*2+result->change*log2((double)12)+result->indel*3;
	printf("  1 avg algn=%8.3f ( - chars - len = %8.3f )\n",
		AdaptiveML(result->match,result->change,result->indel),
		AdaptiveML(result->match,result->change,result->indel)
		-charcost-lencost );
 
	printf(" null-theory= %6.2f = %6.2f(sumlen) + %6.2f(difflen) + %6.2f(chars)  bits\n", nullth, U((float)m+n), DiffCode(m,n), (m+n)*2.0) ;
	pcost = paramcost3(result->match, result->change, result->indel);
	printf("    OneState= %6.2f = %6.2f(pars) + %6.2f(algn) + %6.2f(len)\n",
		OS_messagelength(result), pcost, result->MesLen, lencost) ;
	printf("algn - chars= %6.2f\n", result->MesLen-charcost);

	oddsratio = (float) exp2((double) nullth - rth );
	printf("probability related=%8.4f\n", oddsratio/(1+oddsratio)) ;

	if (m<SCREENHEIGHT && n<SCREENWIDTH) PlotMatrix(a,b,D,m,n,Table) ;

	if (ofile) OutputMatrix(A,B,D,ofile,"OneState R-Theory",Table) ;

} /*plot*/

static SumMatrix(Dfwd,Drev,D,m,n)
	float Dfwd[][MATRIXSIZE], Drev[][MATRIXSIZE], D[][MATRIXSIZE];
	int m, n ;
{
	int i, j ;

	for (i=0 ; i<=m ; i++)
		for (j=0 ; j <= n; j++)
			D[i][j] = Dfwd[i][j] + Drev[i+1][j+1] ;

	scale(D,m,n) ;		
}

OS_densityplot(A, B, P, outfile)
	string *A, *B ;
	OneStatePBlock *P ;
	FILE *outfile ;
{
	float Dfwd[MATRIXSIZE][MATRIXSIZE], Drev[MATRIXSIZE][MATRIXSIZE] ;
	float D[MATRIXSIZE][MATRIXSIZE] ;
	cell result ;

	if ((m < SCREENHEIGHT && n < SCREENWIDTH) || outfile) {
		a = A->sequence ; b = B->sequence ;
		m = A->length ;   n = B->length ;

		DPA(rev, fixed, P, &result, Drev);
		DPA(fwd, fixed, P, &result, Dfwd);
		SumMatrix(Dfwd,Drev,D,m,n) ;
		plot(A,B,D, &result, outfile) ;
	}
}


OS_ProbToCosts(P)
	OneStatePBlock *P ;
{
	costMatch  = -(float)log2((double)P->match) + 2 ;
	costChange = -(float)log2((double)P->change) + (float) log2((double)12) ;
	costDel = costIns  = -(float)log2((double)P->indel) + 3 ;
}


OS_initcell(c) cell *c ; { c->MesLen=infinity; c->match=c->change=c->indel=0; }


static OS_AdaptiveCost(left,diag,top,equalchar)
	cell *left, *diag, *top ;
	bool equalchar ;
{
	costIns = -log2((double)(left->indel+1)/(sum(left)+3)) + 3 ;
	costDel = -log2((double)(top->indel+1)/(sum(top)+3)) + 3 ;
	if (equalchar) 
		costMatch = -log2((double)(diag->match+1)/(sum(diag)+3)) + 2 ;
	else
		costChange = -log2((double)(diag->change+1)/(sum(diag)+3)) + log2((double)12) ;
}

static weightedfreq(current, left, diag, top, MLleft, MLdiag, MLtop,equalchar)
	cell *current, *left, *diag, *top ;
	float MLleft, MLdiag, MLtop ;
	bool equalchar ;   /* true if the current characters in A and B are equal */
{
	float	MLmin, Pleft, Pdiag, Ptop, Psum ;

	MLmin = fmin3(MLleft, MLdiag, MLtop) ;

	Pleft = exp2((double)MLmin-MLleft) ;
	Pdiag = exp2((double)MLmin-MLdiag) ;
	Ptop = exp2((double)MLmin-MLtop) ;
	Psum = Pleft+Pdiag+Ptop ;

	current->match = ((equalchar+diag->match)*Pdiag + left->match*Pleft + top->match*Ptop)/Psum ;
	current->change = (((!equalchar)+diag->change)*Pdiag + left->change*Pleft + top->change*Ptop)/Psum ;
	current->indel = (diag->indel*Pdiag + (left->indel+1)*Pleft + (top->indel+1)*Ptop)/Psum ;
}

OS_ComputeCurrentCell(current,left,diag,top,equalchar)
	cell *current, *left, *diag, *top ;
	bool equalchar ;
{
	float MLvert, MLhori, MLdiag ;

	MLvert = top->MesLen + costDel ;
	MLhori = left->MesLen + costIns ;
	if (equalchar) MLdiag = diag->MesLen + costMatch ;
	else           MLdiag = diag->MesLen + costChange ;

	current->MesLen = -sumlg3(-MLdiag, -MLvert, -MLhori) ;

	weightedfreq(current,left,diag,top,MLhori,MLdiag,MLvert,equalchar);
}

/*
 *	M points to the space to store the matrix of message length
 *	If M is null then it is ignored
 */
static DPA(dirn, mode, P, result, M)
	int   dirn, mode ;
	OneStatePBlock *P ;
	cell *result ;
	float M[][MATRIXSIZE] ;
{
	cell c1[MAXSEQLENGTH], c2[MAXSEQLENGTH], boundary ;
	cell *d1, *d2, *temp ;
	int i, iloop, j, jloop, step, abdry, bbdry ;

	if (mode == fixed) OS_ProbToCosts(P) ;

	d1 = c1 ; d2 = c2 ;
	if (dirn == fwd) { abdry = bbdry = 0; step = 1 ;}
	else             { abdry = m+1; bbdry = n+1 ; step = -1 ;}

	OS_initcell(&boundary) ;
	OS_initcell(&d1[bbdry]) ;
	d1[bbdry].MesLen = 0.0 ;

	i = abdry ;
	j = bbdry ;
	for (jloop = 1 ; jloop <= n; jloop++) {
		j += step ;
		if (mode == adaptive)
			OS_AdaptiveCost(&d1[j-step],&boundary,&boundary,FALSE) ;
		OS_ComputeCurrentCell(&d1[j],&d1[j-step],&boundary,&boundary,FALSE) ;
		if (M) M[i][j] = d1[j].MesLen ;
	}
 
	for (iloop = 1 ; iloop <= m ; iloop++) { /* General Step */
		i += step ;
		j = bbdry ;
		if (mode == adaptive)
			OS_AdaptiveCost(&boundary,&boundary,&d1[j],FALSE) ;
		OS_ComputeCurrentCell(&d2[j],&boundary,&boundary,&d1[j],FALSE) ;
		if (M) M[i][j] = d2[j].MesLen ;

		for (jloop = 1 ; jloop <= n; jloop++) {
			j += step ;
			if (mode == adaptive)
				OS_AdaptiveCost(&d2[j-step],&d1[j-step],&d1[j],a[i]==b[j]) ;
			OS_ComputeCurrentCell(&d2[j],&d2[j-step],&d1[j-step],&d1[j],a[i]==b[j]) ;
			if (M) M[i][j] = d2[j].MesLen ;
		} /* for j */
		swap(d1, d2) ;
	} /* for i */
	*result = d1[n] ;
} /*DPA*/


/*--------------------------------------------------------------------------
 *	Computes the OneState r-theory message length, and return the optimal probs
 *	If mode = adaptive then the first run is an adaptive run to estimate
 *	the starting iteration parameters.   Otherwise the parameters in P
 *	is used for the 1st iterations.  If Outfile is not null (and is a legal
 *	pointer to a file) the density plot will be generated and output to it
 */
OneStateML(A, B, MesgLength, MatrixPart, mode, outfile, P, Accurarcy)
	string *A, *B ;
	OneStatePBlock *P ;
	int mode ;
	FILE *outfile ;
	float Accurarcy, *MesgLength, *MatrixPart ;
{
	float	SumFreq, ML ;
	int		iterations ;
	cell	result ;

	a = A->sequence ; b = B->sequence ;
	m = A->length ;   n = B->length ;

	if (outfile && (m > MATRIXSIZE || n > MATRIXSIZE)) {
		printf("OneState: string length too long, no density plot produced\n") ;
		outfile = NULL ;
	}

	*MesgLength = infinity ;	/* any big number will do */
	iterations = 1 ;

	if (mode == fixed) {
		printf("iteration  0: ",iterations) ; OS_printparameters(P) ;
		putchar('\n') ;
	}

	do {
		ML = *MesgLength ;
		DPA(fwd, mode, P, &result, (float **) NULL);
            
		mode = fixed;
		SumFreq = sum(&result);

		*MesgLength = OS_messagelength(&result) ;
		*MatrixPart = result.MesLen ;
		fchangetoprobability3(result.match,result.change,result.indel,&P->match,&P->change,&P->indel) ;
		printf("iteration %2d: ",iterations) ; OS_printparameters(P) ;
		printf(" ML=%8.3f (%8.3f) SumFq:%7.1f\n", *MesgLength, result.MesLen,SumFreq) ;
		fflush(stdout) ;
		iterations++ ;
	} while (flabs((ML-*MesgLength)/ *MesgLength) > Accurarcy) ;

	printf("iteration %2d:\n",iterations) ;
	OS_densityplot( A, B, P, outfile) ;
}

/*--------------------------------------------------------------------------*/

#define M 1 
#define C 2 
#define I 3
#define D 4

static int
nextstate(P)
	OneStatePBlock	*P ;
{
	float t ;
	int	event ;

	t = (float)(rand() % 100000) / 100000 ; /* 0 <= t < 1 */
	if      ((t -= P->match) < 0)     event = M ;
	else if ((t -= P->change) < 0)    event = C ;
	else if ((t -= (P->indel/2)) < 0) event = I ;
	else                              event = D ;

	return event ;
}

/*
 *	A One-State machine to generate a pair of strings according to
 *	the given set of parameters.
 *	Length refer to the average length of the two strings
 *	The actual generative alignment is also returned
 */
OneStateGen(A, B, length, P, Alignment, Alignlength)
	string *A, *B ;
	OneStatePBlock *P ;
	align  Alignment[] ;
	int    *Alignlength ;
{
	int event, i ;
	int	alen, blen ;
	symbol *s ;
	symbol a[MAXSEQLENGTH], b[MAXSEQLENGTH] ;

	*Alignlength = alen = blen = 0 ;
	while ((alen + blen) < 2*length) {
		event = nextstate(P) ;
		++(*Alignlength) ;
		if (event == M) {
			a[++alen] = b[++blen] = randomsymbol(&DNASymbol) ;
			Alignment[*Alignlength].a = alen ;
			Alignment[*Alignlength].b = blen ;
		} else if (event == C) {
			randomsymbolpair(&DNASymbol, &a[++alen], &b[++blen] ) ;
			Alignment[*Alignlength].a = alen ;
			Alignment[*Alignlength].b = blen ;
		} else if (event == I) {
			b[++blen] = randomsymbol(&DNASymbol) ;
			Alignment[*Alignlength].a = 0 ;
			Alignment[*Alignlength].b = blen ;
		} else {
			a[++alen] = randomsymbol(&DNASymbol) ;
			Alignment[*Alignlength].a = alen ;
			Alignment[*Alignlength].b = 0 ;
		}

		if (alen > MAXSEQLENGTH || blen > MAXSEQLENGTH) {
			printf("OneStateGen : string too long\n") ;
			exit(99) ;
		}
	}
	s = A->sequence = (symbol *) malloc((unsigned)(alen+1)*sizeof(symbol)) ;
	A->length = alen ;
	s[0] = NULLSymbol ;
	for (i = 1; i <= alen; i++) s[i] = a[i] ;

	s = B->sequence = (symbol *) malloc((unsigned)(blen+1)*sizeof(symbol)) ;
	B->length = blen ;
	s[0] = NULLSymbol ;
	for (i = 1; i <= blen; i++) s[i] = b[i] ;
}
