/*
 *  Copyright  (c)  L.Allison, C.S.Wallace, C.N.Yee.
 *     Dept. Computer Science, Monash University, Australia 3168,    May 1991.
 *  1. These routines may only be used for non-commercial, non-classified,
 *     research purposes.
 *  2. No liability of any kind is accepted for their use.
 *  3. As a courtesy, please inform us of any interesting uses of the routines,
 *     of any bugs that you find in them and of how to correct
 *     such bugs, if known.
 *  4. Do not delete this copyright notice.
 *  5. Please include the following references in any publication based wholly
 *     or in part on the use of these routines or on the ideas contained in them:
 *   . Allison L. & C.N.Yee.
 *        Minimum Message Length Encoding and the Comparison of Macro-Molecules.
 *        Bull. Math. Bio. 52(3) 431-453 1990
 *   . Allison L., C.S.Wallace & C.N.Yee. Finite-State Models in the Alignment
 *        of Macromolecules. J. Mol. Evol. (1992) 35:77-89
 *  6. You may distribute the routines to others if and only if
 *     they agree to accept and maintain these conditions and no fee is charged.
 *  7. There is no rule 7.
 */
/*
 *	"1StateR0.c" --
 *		computes a best R-model alignment usign Hirschberg's recursive algorithm
 *
 *	Adapted from D. S. Hirschberg
 *  	 'A linear space algorithm for computing maximal common subsequences.'
 *  	 CACM V18(6) p341-343 June 1975     (ie. LCS)
 *
 * Entries :
 *
 * RAlignString( AStr, BStr, Alignment, AlignLen, Matches, DiffFunct )
 *		string	*AStr, *BStr ;
 *		int	*AlignLen, *Matches; 
 *		align	Alignment[] ;
 *		int	(*DiffFunct)() ;
 *
 * RAlignSequence(ASeq, BSseq, Alen, Blen, Alignment, AlignLen, P, DiffFunct )
 *
 * R0MesgLength(Alignment, Alignlength, A, B, Msglength, Table)
 *  	align	Alignment[] ;
 *		int	Alignlength ;
 *		symbol	A[], B[] ;
 *		AlphabetTable	*Table ;
 *		float	*Msglength ;
 *
 *		computes the message length of an R-alignment using the
 *		(=, <>, indel) adaptive coding
 *
 * 
 * ROptAlignString(A,B,Alignment,AlignLen,msglength,Matches,Niteration,P,Table)
 *	 	string	*A, *B ;
 * 		int	*AlignLen, *Matches, *Niteration ;
 * 		align	Alignment[] ;
 *		OneStatePBlock *P ;
 * 		float	*msglength ; 
 * 		AlphabetTable	*Table ;
 *
 *		iterates on the parameters until an fixed point is reached
 *		the resultant probabolities and number of iteration is returned
 */
#include <stdio.h>
#include "macro.h"
#include "Strings.h"
#include "rtheory.h"

static symbol 	*A, *B ;
static align	*AlignStr ;
static int	AlignLength, NMatches ;

static float	l1[MAXSEQLENGTH+2], l2[MAXSEQLENGTH+2] ;
static float	(*Diff)() ;

/***************************************************************************/
static copy(s1, s2, lo, hi)
	float	s1[], s2[] ;
	int	lo, hi ;
{
	int	i ;

	for (i = lo; i <= hi ; i++) s1[i] = s2[i] ;
}

static printarray(s,lo,hi)
	float s[] ;
	int	lo, hi ;
{
	int	i ;

	for (i = lo; i <= hi ; i++) printf("%5.1f ",s[i]) ;
	putchar('\n') ;
}


static code(i, j )
	int	i, j ;
{
	AlignStr[++AlignLength].a = i ;
	AlignStr[AlignLength].b = j ;
	if (i != 0 && j != 0 && A[i] == B[j]) NMatches++ ;
}

		
	float c[2][MAXSEQLENGTH+2];
static
AlgB( loA, hiA, loB, hiB, D, l)
	int	loA, hiA, loB, hiB, D ;
	float	l[] ;
{
	int	i, j ;
	float *cost0, *cost1, *temp ;
		
	cost1 = c[1] ;
	cost0 = c[0] ;

	switch (D) {
	case fwd:
		cost1[loB-1] = 0.0 ; /* (*Diff)(NULLSymbol, NULLSymbol);*/
		for (j = loB; j <= hiB ; j++)
			cost1[j] = cost1[j-1] + (*Diff)(NULLSymbol, B[j]);
		for (i = loA; i <= hiA; i++) {
			swap(cost0, cost1) ;
			cost1[loB-1] = cost0[loB-1] + (*Diff)(A[i], NULLSymbol);
			for (j=loB ; j <= hiB; j++)
				cost1[j]= fmin3(cost0[j-1]+(*Diff)(A[i],B[j]),
						cost0[j]  +(*Diff)(A[i], NULLSymbol),
						cost1[j-1]+(*Diff)(NULLSymbol, B[j])) ;
		} ;
		break ;
	case rev:
 		cost1[hiB+1] = 0.0; /* (*Diff)(NULLSymbol,NULLSymbol);*/
 		for (j = hiB ; j >= loB ; j--) 
			cost1[j] = cost1[j+1] + (*Diff)(NULLSymbol, B[j]);
		for (i = hiA ; i >= loA; i--) {
			swap(cost0,cost1) ;
			cost1[hiB+1] = cost0[hiB+1] + (*Diff)(A[i], NULLSymbol) ;
			for (j = hiB; j >= loB; j --)
				cost1[j]= fmin3(cost0[j+1]+(*Diff)(A[i], B[j]),
					cost0[j] + (*Diff)(A[i], NULLSymbol),
					cost1[j+1] + (*Diff)(NULLSymbol, B[j])) ;
		}
		break ;
	} /*case*/;

	copy(l, cost1, loB-1, hiB+1) ;
}   /*AlgB*/;

static
AlgC(mincost, loA, hiA, loB, hiB)
	float	*mincost ;
	int	loA, hiA, loB, hiB ;
{
	int	midA, midB, i, j, k ;
	float	dontcare, skipB ;

	debug(printf("AlgC %d %d %d %d\n", loA, hiA, loB, hiB) ;) 
	if (hiB < loB)    /*B is empty*/
		for (i=loA ; i <= hiA; i++) code( i, NULLSymbol ) ;

	else /*hiB >=loB*/ if (hiA == loA)  /* |A| = 1 */ {
		k=loB-1;
		skipB = 0.0 ;
		for (j=loB; j <= hiB; j++) skipB += (*Diff)(NULLSymbol, B[j]) ;
		*mincost = (*Diff)(A[loA], NULLSymbol) + skipB;
		for (j = loB; j <= hiB; j++)
			if ((skipB - (*Diff)(NULLSymbol,B[j]) + (*Diff)(A[loA],B[j])) < *mincost ) {
				k = j ;
				*mincost = skipB - (*Diff)(NULLSymbol,B[j]) + (*Diff)(A[loA],B[j]) ;
			}
 		for (j= loB; j <= k-1; j++) code(NULLSymbol, j);
		if (k < loB)  code( loA, NULLSymbol ) ;
 		else /*k>=loB*/ code( loA, k );
		for (j= k+1; j <= hiB; j++) code(NULLSymbol, j) ;
	} /*hiA=loA*/
      
	else { /* hiA > loA    so divide and conquer*/ 
		midA = (int) ((loA + hiA) / 2) ;     /* loA <= midA < hiA */
		AlgB(   loA, midA, loB, hiB, fwd, l1);
		AlgB(midA+1,  hiA, loB, hiB, rev, l2);
		debug(
		if (loB == 1 ) {
			printarray(l1,loB-1,hiB) ; printarray(l2,loB,hiB+1) ; }
		)
		midB = loB-1;
		*mincost = l1[loB-1]+l2[loB];
		for (j = loB ; j <= hiB ; j++)
			if (l1[j]+l2[j+1] < *mincost) {
				midB = j;
				*mincost = l1[j]+l2[j+ 1] ;
			}
		AlgC(&dontcare,    loA, midA, loB,   midB) ;
		AlgC(&dontcare, midA+1,  hiA, midB+1, hiB) ;
	}
	debug(printf("    AlgC returns %d %d %d %d\n", loA, hiA, loB, hiB) ;)
} /*AlgC*/;


RAlignSequence(Aseq, Bseq, Alen, Blen, Alignment, AlignLen, Matches, DiffFunct)
	symbol	*Aseq, *Bseq ;
	int	*AlignLen, *Matches; 
	align	Alignment[] ;
	float	(*DiffFunct)() ;
{
	float	mincost ;

	A = Aseq ; B = Bseq ;
	AlignStr = Alignment ;
	Diff = DiffFunct ;
	NMatches = AlignLength = 0 ;

	AlgC(&mincost, 1, Alen, 1, Blen) ;

	*AlignLen = AlignLength ;
	*Matches = NMatches ;
}

RAlignString(A, B, Alignment, AlignLen, Matches, DiffFunct )
	string	*A, *B ;
	int	*AlignLen, *Matches; 
	align	Alignment[] ;
	float	(*DiffFunct)() ;
{
	RAlignSequence(A->sequence, B->sequence, A->length, B->length, Alignment, AlignLen, Matches, DiffFunct) ;
}

/*----------------------------------------------------------------------------*/

#define MAXITERATION 15

typedef struct {
	float	cost ;
	float	PMatches, PChanges; 
	} ihist ;

ihist iterationhistory[MAXITERATION] ;
static int	iterations ;

/*
 *	maintains iteration history and detect iteration cycle
 */
static bool
repeat(cost,PMatches,PChanges)
	float	cost,PMatches,PChanges ;
{
	int i ;

	for ( i = iterations-1 ; i >= 0 ; i--) {
		if ((PMatches == iterationhistory[i].PMatches)
			&& (PChanges == iterationhistory[i].PChanges)) {
			break ;
		}
	}
	if (i < 0) {/* not found, register this iteration */
		iterationhistory[iterations].cost = cost ; 
		iterationhistory[iterations].PMatches = PMatches ; 
		iterationhistory[iterations].PChanges = PChanges ; 
		return FALSE ;
	} else
		return TRUE ;
}

static
lookformin(where,PMatches,PChanges)
	int	*where ;
	float *PMatches, *PChanges ;
{
	int i ;
	*where = iterations - 1 ;
	for ( i = iterations-2 ; i >= 0; i-- )
		if (iterationhistory[i].cost < iterationhistory[*where].cost)
			*where = i ;
	*PMatches = iterationhistory[*where].PMatches ;
	*PChanges = iterationhistory[*where].PChanges ;
}

static float	CostMatch, CostChange, CostInsDel ;

static float
InitRCostFunction(P)
	OneStatePBlock *P ;
{
	CostMatch = (float)- log2((double)P->match) + 2 ;
	CostChange = (float)-log2((double)P->change) + log2((double)12) ;
	CostInsDel = (float)- log2((double)(P->indel)) + 3 ;
}

static float
RCostFunct(a, b)
	symbol a, b ;
{
	if (a == b) return CostMatch ;
	if (a == NULLSymbol || b == NULLSymbol) return CostInsDel ;
	else return CostChange ;
}

/*
 *	iterates on the parameters until an fixed point is reached
 */
ROptAlignString(A,B,Alignment,AlignLen,msglength,Niteration,P,Table)
	string	*A, *B ;
	int	*AlignLen, *Niteration ;
	align	Alignment[] ;
	float	*msglength ;
	OneStatePBlock *P ;
	AlphabetTable	*Table ;
{
	int	where, M ;

	iterations = 0 ;
	while (TRUE) {
		InitRCostFunction(P) ;
		RAlignString(A, B, Alignment, AlignLen, &M, RCostFunct) ;

		R0MesgLength(Alignment,*AlignLen,A->sequence,B->sequence,msglength,Table);
		OS_computeparameters(Alignment,*AlignLen,A->sequence,B->sequence,P) ;
		printf("   r-0 :%1.3f %1.3f %1.3f ML=%5.2f AlignLen : %4d\n",P->match,P->change,P->indel,*msglength,*AlignLen);
		fflush(stdout) ;

		debug(
		PrintAlignment(Alignment,*AlignLen,A->sequence,B->sequence,Table) ;
		)
		if (repeat(*msglength,P->match,P->change)) {
			break ;
		}
		iterations++ ;
	}
	lookformin(&where,&P->match,&P->change) ;
	if (where != iterations-1) {
		P->indel = 1 - P->match - P->change ;
		InitRCostFunction(P) ;
		RAlignString(A, B, Alignment, AlignLen, &M, RCostFunct) ;
		R0MesgLength(Alignment,*AlignLen,A->sequence,B->sequence,msglength,Table);
	}
	*Niteration = iterations ;
}

/**********************************************************************/
/*
 *	compute the message length of an alignment using the
 *	(=,<>,Indel) coding
 */
R0MesgLength(Alignment, Alignlength, A, B, Msglength, Table)
	align	Alignment[] ;
	int	Alignlength ;
	symbol	A[], B[] ;
	AlphabetTable	*Table ;
	float	*Msglength ;
{
	int	i, NInsDel, NMatches, NChanges ; 
	float	msglength, loga, logab ;

	NInsDel = NMatches = NChanges = 0 ; 
	msglength = 0.0 ;

	loga = (float) log2((double)Table->size) ;
	logab = (float) log2((double)(Table->size * (Table->size-1))) ;

	for ( i = 1; i <= Alignlength; i++) {
		if ((Alignment[i].a == 0) || (Alignment[i].b == 0 )) {
			msglength += (float) -log2((double)++NInsDel/(i+2)) + 1 + loga ;
		} else if (A[Alignment[i].a] == B[Alignment[i].b]) {
			msglength += (float)-log2((double)++NMatches/(i+2)) + loga ;
		} else {
			msglength += (float)-log2((double)++NChanges/(i+2)) + logab ;
		}
	}
	*Msglength = msglength ;
}

/*
 *	this routine computes the relative frequencies of
 *	Matches, Chnages and InsDel	of an One State R0-alignment
 */
OS_computeparameters(Alignment, Alignlength, A, B, P )
	align	Alignment[] ;
	int	Alignlength ;
	symbol	A[], B[] ;
	OneStatePBlock	*P ;
{
	int	i, NInsDel, NMatches, NChanges ; 

	NInsDel = NMatches = NChanges = 0 ; 

	for ( i = 1; i <= Alignlength; i++) {
		if ((Alignment[i].a == 0) || (Alignment[i].b == 0 )) NInsDel++ ;
		else if (A[Alignment[i].a] == B[Alignment[i].b]) NMatches++ ;
		else NChanges ++ ;
	}
	changetoprobability3(NMatches,NInsDel,NChanges,&P->match,&P->indel,&P->change) ;
}

/*----------------------------------------------------------------------------*/

struct element {
		int M, C, ID ;
		float	ways ;
	} ;

static float
cost(c)
	struct element *c ;
{
	return c->M * CostMatch + c->C * CostChange + c->ID * CostInsDel ;
}

static bool
close(a, b)
	float	a, b ;
{
	if (a != 0 || b !=0) return (flabs((a-b)/(a+b)) < 0.0001) ;
	return TRUE ;
}

/*
 *	returns log2 the number of ways of reaching
 *	a particular optimal alignment
 */
float
RCount(AStr, BStr, P)
	string	*AStr, *BStr ;
	OneStatePBlock *P ;
{
	symbol *A, *B ;
	int	Alength, Blength, i, j ;
	struct element l1[MAXSEQLENGTH], l2[MAXSEQLENGTH], 
			left, top, diag, *c1, *c2, *temp ;
	float leftcost,topcost,diagcost,mincost ;
	float ways ;

	InitRCostFunction(P) ;

	A = AStr->sequence ; B = BStr->sequence ;
	Alength = AStr->length ; Blength = BStr->length ;
	c1 = l1 ; c2 = l2 ;

	/* boundary conditions */
	for ( j = 0 ; j <= Blength ; j++) {
		c1[j].M = c1[j].C = 0 ;
		c1[j].ID = j ;
		c1[j].ways = 0 ;
	}

	for ( i = 1 ; i <= Alength; i++) {
		c2[0].M = c2[0].C = 0 ;
		c2[0].ID = i ;
		c2[0].ways = 0 ;

		for (j = 1 ; j <= Blength; j++) {
			left.M = c2[j-1].M ; left.C = c2[j-1].C; left.ID = c2[j-1].ID+1;
			top.M = c1[j].M ; top.C = c1[j].C; top.ID = c1[j].ID+1;
			if (A[i] == B[j]) {
				diag.M = c1[j-1].M+1 ; diag.C = c1[j-1].C; diag.ID = c1[j-1].ID;
			} else {
				diag.M = c1[j-1].M ; diag.C = c1[j-1].C+1; diag.ID = c1[j-1].ID;
			}

			leftcost = cost(&left) ;
			topcost = cost(&top) ;
			diagcost = cost(&diag) ;

			mincost = fmin3(leftcost,topcost,diagcost) ;
			ways = -1999 ;
			if (close(mincost, leftcost)) {
				c2[j] = left ;
				ways = sumlg(ways, c2[j-1].ways) ;
			}
			if (close(mincost, diagcost)) {
				c2[j] = diag ;
				ways = sumlg(ways,c1[j-1].ways) ;
			}
			if (close(mincost, topcost)) {
				c2[j] = top ;
				ways = sumlg(ways,c1[j].ways) ;
			}
			c2[j].ways = ways ;
		}
		swap(c1,c2) ;
	}
	debug(
	)
	printf("   Rcount - cost %1.2f  ways %5.3f\n",mincost, c1[Blength].ways) ;
	return c1[Blength].ways ;
}

