
/*
 *  Copyright  (c)  L.Allison, C.S.Wallace, C.N.Yee.
 *     Dept. Computer Science, Monash University, Australia 3168,    May 1991.
 *  1. These routines may only be used for non-commercial, non-classified,
 *     research purposes.
 *  2. No liability of any kind is accepted for their use.
 *  3. As a courtesy, please inform us of any interesting uses of the routines,
 *     of any bugs that you find in them and of how to correct
 *     such bugs, if known.
 *  4. Do not delete this copyright notice.
 *  5. Please include the following references in any publication based wholly
 *     or in part on the use of these routines or on the ideas contained in them:
 *   . Allison L. & C.N.Yee.
 *        Minimum Message Length Encoding and the Comparison of Macro-Molecules.
 *        Bull. Math. Bio. 52(3) 431-453 1990
 *   . Allison L., C.S.Wallace & C.N.Yee. Finite-State Models in the Alignment
 *        of Macromolecules. J. Mol. Evol. (1992) 35:77-89
 *  6. You may distribute the routines to others if and only if
 *     they agree to accept and maintain these conditions and no fee is charged.
 *  7. There is no rule 7.
 */
/*	"3State.c"  prefix TS
 *
 *	Compute the prob of 2 strings being related under 3 state R-theory
 *
 */
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "macro.h"
#include "Strings.h"
#include "rtheory.h"

float sumD(), sumH(), sumV() ; /* forward declarations */

/*
 *	some global variables
 */
static symbol   *a, *b ;
static int m,n ;

float costDM, costDC, costDH, costDV ;
float costHM, costHC, costHH, costHV ;
float costVM, costVC, costVH, costVV ;

/*---------------------------------------------------------------------------
 *	Computes the corresponding edit cost from a set of arcs probabilities
 */
TS_ProbToCosts(P,CostDM,CostDC,CostDH,CostDV,CostHM,CostHC,CostHH,CostHV,CostVM,CostVC,CostVH,CostVV)
	ThreeStatePBlock *P ;
	float *CostDM,*CostDC,*CostDH,*CostDV,*CostHM,*CostHC,*CostHH,*CostHV,*CostVM,*CostVC,*CostVH,*CostVV ;
{
	float log12 ;

	log12 = (float) log2((double)(12)) ;

	*CostDM  = (float)log2((double)P->DM) - 2 ;
	*CostDC  = (float)log2((double)P->DC) - log12 ;
	*CostDH  = (float)log2((double)P->DH) - 2 ;
	*CostDV  = (float)log2((double)P->DV) - 2 ;

	*CostHM  = (float)log2((double)P->HM) - 2 ;
	*CostHC  = (float)log2((double)P->HC) - log12 ;
	*CostHH  = (float)log2((double)P->HH) - 2 ;
	*CostHV  = (float)log2((double)P->HV) - 2 ;

	*CostVM  = (float)log2((double)P->VM) - 2 ;
	*CostVC  = (float)log2((double)P->VC) - log12 ;
	*CostVH  = (float)log2((double)P->VH) - 2 ;
	*CostVV  = (float)log2((double)P->VV) - 2 ;
}

/*----------------------------------------------------------------------------
 *	The following two routines is a O(n^4) brute force algo to compute the 
 *	density plot for 3 state model	--- for debugging and comparison only.
 *	The fwd matrix is not neccessary, but will save half the time
 */
static DensityDPA(start, result, m1, n1)
	int n1, m1 ; /* the mid starting point */
	TSpcell *start ; /* the initial condition */
	float *result ;
{
	TSpcell e1[MAXSEQLENGTH], e2[MAXSEQLENGTH] ;
	TSpcell *d1, *d2, *temp ;
	int i, j ;

	d1 = e1 ;
	d2 = e2 ;
	/* Boundary Conditions */
	d1[n1] = *start ;
      
	for (j = n1+1 ; j <= n; j++) {
		if (j == n1+1)
			d1[j].H = sumlg3(d1[n1].D+costDH, d1[n1].H+costHH, d1[n1].V+costVH);
		else
			d1[j].H = d1[j-1].H + costHH ;

		d1[j].D = d1[j].V = BIG ;
	}
 
	for (i= m1+1 ; i<= m ; i++) { /* General Step */
		if (i == m1+1) 
			d2[n1].V = sumlg3(d1[n1].D+costDV,d1[n1].H+costHV,d1[n1].V+costVV);
		else
			d2[n1].V = d1[n1].V + costVV ;
		d2[n1].D = d2[n1].H = BIG ;

		for (j = n1+1; j <= n; j++) {
			if (a[i] == b[j]) d2[j].D= sumlg3(d1[j-1].D+costDM,d1[j-1].H+costHM,d1[j-1].V+costVM);
			else d2[j].D = sumlg3(d1[j-1].D+costDC,d1[j-1].H+costHC,d1[j-1].V+costVC);
			d2[j].H =sumlg3(d2[j-1].D+costDH,d2[j-1].H+costHH,d2[j-1].V+costVH);
			d2[j].V = sumlg3(d1[j].D+costDV,d1[j].H+costHV, d1[j].V+costVV) ;
		} /* for j */
		swap(d1, d2) ;
	} /* for i */
	*result = -sumlg3(d1[n].D,d1[n].H,d1[n].V) ;
} /*DensityDPA*/

static
BruteForceDensity(Fwd,Rslt,P)
	TSpcell Fwd[][MATRIXSIZE] ;
	float Rslt[][MATRIXSIZE] ;
	ThreeStatePBlock *P ;
{
	int	 i, j ;

	TS_ProbToCosts(P,&costDM,&costDC,&costDH,&costDV,&costHM,&costHC,&costHH,&costHV,&costVM,&costVC,&costVH,&costVV) ;

	for (i = 0; i <= m; i++)
		for (j = 0; j <= n ; j++)
			DensityDPA(&Fwd[i][j],&Rslt[i][j],i,j) ;
}

/*------------------------------------------------------------------------*/

static float MessageLength(C)
	TScell *C ;
{
	float	t1,t2,t3,t4,t5 ;
	t1 = paramcost4(C->DM, C->DC, C->DH, C->DV) ;
	t2 = paramcost4(C->HM, C->HC, C->HH, C->HV) ;
	t3 = paramcost4(C->VM, C->VC, C->VV, C->VH) ;
	t4 = -sumlg3(C->D, C->H, C->V) ;
	t5 =  U(sumH(C) + sumV(C) + sumD(C)) ;
	/*
	printf(" %4.2f(parD) %4.2f(parH) %4.2f(parV) %4.2f(mat) %4.2f(U)\n"
		,t1 ,t2 ,t3 ,t4 ,t5) ;
	*/
	return (t1+t2+t3+t4+t5) ;
}

/*
 *	when states H & V are identical
 */
static float MessageLength2(C)
	TScell *C ;
{
	float	t1,t2,t3,t4 ;

	t1 = paramcost3(C->DM, C->DC, C->DH+C->DV) ;
	t2 = paramcost4(C->HM+C->VM, C->HC+C->VC, C->HH+C->VV, C->HV+C->VH) ;
	t3 = -sumlg3(C->D, C->H, C->V) ;
	t4 = U(sumH(C) + sumV(C) + sumD(C)) ;
	/*
	printf(" %4.2f(parD) %4.2f(parHV) %4.2f(mat) %4.2f(U)\n" ,t1 ,t2 ,t3 ,t4) ;
	*/
	return (float)(t1 + t2 + t3 + t4) ;
}


/*--------------------------------------------------------------------------
 *	Some general utility routines
 */
static float sumD(C )
	TScell *C ;
{ return C->DM + C->DC + C->DH + C->DV ; }

static float sumH(C )
	TScell *C ;
{ return C->HM + C->HC + C->HH + C->HV ; }

static float sumV(C )
	TScell *C ;
{ return C->VM + C->VC + C->VH + C->VV ; }

static copy(p, q)
	TScell *q ; TSpcell *p ;
{ p->D = q->D ; p->H = q->H ; p->V = q->V ; }

TS_initcell(p)
	TScell *p ;
{
	p->D =  p->H =  p->V = BIG ;
	p->DM = p->DC = p->DH = p->DV = 0.0 ;
	p->HM = p->HC = p->HH = p->HV = 0.0 ;
	p->VM = p->VC = p->VH = p->VV = 0.0 ;
}
/*-------------------------------------------------------------------------*/

#ifdef debug
static TS_printcell(i,j,c)
	int i,j;
	TScell *c ;
{
	printf("(%d,%d) %4.2f %4.2f %4.2f %4.2f %4.2f %4.2f %4.2f %4.2f %4.2f %4.2f %4.2f %4.2f %4.2f %4.2f %4.2f\n",
	i,j,c->D,c->H,c->V,c->DM,c->DC,c->DH,c->DV,c->HM,c->HC,c->HH,c->HV,c->VM,c->VC,c->VH,c->VV) ;
}
#endif

TS_printparameters(P)
	ThreeStatePBlock *P ;
{
	printf("       M     C     H     V\n") ;
	printf("    D %5.3f %5.3f %5.3f %5.3f\n",P->DM,P->DC,P->DH,P->DV) ;
	printf("    H %5.3f %5.3f %5.3f %5.3f\n",P->HM,P->HC,P->HH,P->HV) ;
	printf("    V %5.3f %5.3f %5.3f %5.3f\n",P->VM,P->VC,P->VH,P->VV) ;
}

TS_printparameters2(P,result,MesgLength)
	ThreeStatePBlock *P ;
	float MesgLength ;
	TScell *result ;
{
	printf("               M     C     H     V    sumfq \n") ;
	printf("    D %5.4f %5.3f %5.3f %5.3f %5.3f %5.1f\n",P->D,P->DM,P->DC,P->DH,P->DV, sumD(result)) ;
	printf("    H %5.4f %5.3f %5.3f %5.3f %5.3f %5.1f\n",P->H,P->HM,P->HC,P->HH,P->HV, sumH(result)) ;
	printf("    V %5.4f %5.3f %5.3f %5.3f %5.3f %5.1f\n",P->V,P->VM,P->VC,P->VH,P->VV, sumV(result)) ;
	printf("    ML=%10.4f (matrix part=%10.4f)\n", MesgLength,-sumlg3(result->D,result->H,result->V)) ;

}

TS_statistics(P,result,MesgLength)
	ThreeStatePBlock *P ;
	float *MesgLength ;
	TScell *result ;
{
	fchangetoprobability4(result->DM,result->DC,result->DH,result->DV,&P->DM,&P->DC,&P->DH,&P->DV) ;
	fchangetoprobability4(result->HM,result->HC,result->HH,result->HV,&P->HM,&P->HC,&P->HH,&P->HV) ;
	fchangetoprobability4(result->VM,result->VC,result->VH,result->VV,&P->VM,&P->VC,&P->VH,&P->VV) ;

	/*
	   the fractional time that the machine is in the various states
	   this is deternimed by the no. of times the machine enters
	   the various state,  according to the weighted frequency counter
	*/
	fchangetoprobability3(
		result->DM+result->HM+result->VM+result->DC+result->HC+result->VC,
		result->DH+result->HH+result->VH,
		result->DV+result->HV+result->VV,
		&P->D,&P->H,&P->V) ;

	*MesgLength = MessageLength(result) ;
}

/*
 *	this one make states 2 and 3 symmetric(equal)
 */
TS_statistics2(P,result,MesgLength)
	ThreeStatePBlock *P ;
	float *MesgLength ;
	TScell *result ;
{
	fchangetoprobability3(result->DM,result->DC,result->DH+result->DV,&P->DM,&P->DC,&P->DH) ;
	P->DH = P->DV = P->DH/2 ;

	fchangetoprobability4(result->HM+result->VM, result->HC+result->VC, result->HH+result->VV,result->HV+result->VH, &P->HM,&P->HC,&P->HH,&P->HV) ;
	P->VM = P->HM ; P->VC = P->HC ;
	P->VV = P->HH ; P->VH = P->HV ;

	/* probs of each state according to weighted count */
	fchangetoprobability2(result->DM+result->HM+result->VM+result->DC+result->HC+result->VC, result->DH+result->HH+result->VH + result->DV+result->HV+result->VV,&P->D,&P->H) ;
	P->V = P->H ;

	*MesgLength = MessageLength2(result) ;
}

/*------------------------------------------------------------------------*/
static printmatrix(M)
	float M[][MATRIXSIZE] ;
{
	int i, j ;

	printf("       ") ;
	for ( j = 1 ; j <= n ; j++) printf("  %c ",SymbTable->Symbol[b[j]]) ;
	putchar('\n') ;
	for ( i = 0 ; i <= m ; i++) {
		if ( i != 0) printf(" %c ",SymbTable->Symbol[a[i]]) ;
		else printf("   ") ;
		for ( j = 0 ; j <= n ; j++)
			printf(" %3d",(int) (M[i][j]+0.5)) ;
		putchar('\n') ;
	}
}

/*
 *	combine Fwd and Rev
 */
static combine(Fwd,Rev,DMatrix)
	TSpcell Fwd[][MATRIXSIZE],Rev[][MATRIXSIZE] ;
	float DMatrix[][MATRIXSIZE] ;
{
	int i, j ;

	for ( i = 0 ; i <= m ; i++)
		for ( j = 0 ; j <= n ; j++) {
			DMatrix[i][j] = -sumlg3(Fwd[i][j].D+Rev[i+1][j+1].D,
			 		Fwd[i][j].H+Rev[i+1][j+1].H,
			 		Fwd[i][j].V+Rev[i+1][j+1].V);
		}
}

/*
 *	Do a forward run then a reverse run using the parameters P
 *	then combine the fwd and rev result output into OFILe the
 *	resulting density plot
 */
static DensityPlot(A,B,P,ofile)
	string *A, *B ;
	ThreeStatePBlock *P ;
	FILE *ofile ;
{
	TSpcell Fwd[MATRIXSIZE][MATRIXSIZE], Rev[MATRIXSIZE][MATRIXSIZE] ;
	float	DMatrix[MATRIXSIZE][MATRIXSIZE] ;
	int i, j ;
	TScell result ;

	if (ofile || (m < SCREENHEIGHT && n < SCREENWIDTH)) {

		DPA(fixed,P,&result, Fwd) ;
		DPArev(P,Rev) ;

		combine(Fwd,Rev,DMatrix) ;
		scale(DMatrix,m,n) ;
		if (m < SCREENHEIGHT && n < SCREENWIDTH) PlotMatrix(a,b,DMatrix,m,n,SymbTable) ;

		if (ofile) OutputMatrix(A,B,DMatrix,ofile,"Three-State R-Theory",SymbTable) ;

	}
}

/*--------------------------------------------------------------------------
 *	Computes the 3 states R-theory message length, and return the optimal probs
 */
ThreeStateMesLength(A,B,MesgLength,MatrixPart,mode,outfile,P,Accurarcy,Symmetric)
	string *A, *B ;
	FILE *outfile ;	/* if not NULL the density plot will be generated and
						output to it */
	int mode ;	/* if equal fixeed then use the supplied parameters */
				/* else use adaptive coding for the first run to guess a set
				   of good parameters */
	int Symmetric ;
	float *MesgLength,*MatrixPart ;
	float Accurarcy ; /* the stopping accurarcy */
	ThreeStatePBlock *P ;
{
	float OldML ;
	int iterations ;
	TScell result ;

	char filename[100] ;

	a = A->sequence ; b = B->sequence ;
	m = A->length ;   n = B->length ;

	if (outfile && (m > MATRIXSIZE || n > MATRIXSIZE)) {
		printf("ThreeStateML() : wraning : string too long, no density plot produced\n") ;
		outfile = (FILE *) NULL ;
	}

	iterations = 1 ;
	if (mode==adaptive) P->D = P->H = P->V = 1.0/3 ;
	else {
		printf("Iteration  0:\n") ;
		TS_printparameters(P) ;
	}

	*MesgLength = infinity ;
	/* 
	 *	Matrix part always decreases, however the total message length (including par costs)
	 *	may increase slightly. We should stop the iteration then
	 */
	do {
		OldML = *MesgLength ;
		DPA(mode,P,&result,(TSpcell *) NULL) ;

		if (Symmetric) TS_statistics2(P,&result,MesgLength) ;
		else TS_statistics(P,&result,MesgLength) ;

		printf("  Iteration %2d :\n",iterations) ;
		TS_printparameters2(P,&result,*MesgLength) ;
		fflush(stdout) ;

		mode = fixed;
		iterations++ ;
	} while (((OldML-*MesgLength)/ *MesgLength) > Accurarcy) ;

	DensityPlot(A,B,P,outfile) ;

	*MatrixPart = -sumlg3(result.D,result.H,result.V) ;
}

/*------------------------------------------------------------------------*/
/*
 *	a count of 1 is distrubuted to the three counters S1, S2 and S3 
 *	weighted by the message length (log probabilities) L1, L2 and L3.
 */
static scaledprobability3(L1,L2,L3,P1,P2,P3,Psum)
	float L1,L2,L3,*P1,*P2,*P3,*Psum ;
{
	float Lmin ;

	Lmin = fmin3(L1,L2,L3) ;
	*P1 = exp2((double)Lmin-L1) ;
	*P2 = exp2((double)Lmin-L2) ;
	*P3 = exp2((double)Lmin-L3) ;
	*Psum = *P1 + *P2 + *P3 ;
}

static TS_weightedincrement(L1, L2, L3, S1, S2, S3)
	float	L1, L2, L3 ;
	float	*S1, *S2, *S3 ;
{
	float Lmin, P1, P2, P3, Psum ;

	scaledprobability3(L1,L2,L3,&P1,&P2,&P3,&Psum) ;
	*S1 += P1/Psum ;
	*S2 += P2/Psum ;
	*S3 += P3/Psum ;
}

/*
 *	Compute the average of the counters in left, diag and top weighted
 *	by the probabilities of enter from the corresponding directions
 *	
 */
static TS_weightedaverage(current, left, diag, top) 
	TScell *current, *left, *diag, *top ;
{
	float	Pleft, Pdiag, Ptop, Psum ;

	scaledprobability3(-current->H,-current->D,-current->V,&Pleft,&Pdiag,&Ptop,&Psum) ;
	current->DM = (diag->DM*Pdiag + left->DM*Pleft + top->DM*Ptop)/Psum ;
	current->DC = (diag->DC*Pdiag + left->DC*Pleft + top->DC*Ptop)/Psum ;
	current->DH = (diag->DH*Pdiag + left->DH*Pleft + top->DH*Ptop)/Psum ;
	current->DV = (diag->DV*Pdiag + left->DV*Pleft + top->DV*Ptop)/Psum ;

	current->HM = (diag->HM*Pdiag + left->HM*Pleft + top->HM*Ptop)/Psum ;
	current->HC = (diag->HC*Pdiag + left->HC*Pleft + top->HC*Ptop)/Psum ;
	current->HH = (diag->HH*Pdiag + left->HH*Pleft + top->HH*Ptop)/Psum ;
	current->HV = (diag->HV*Pdiag + left->HV*Pleft + top->HV*Ptop)/Psum ;

	current->VM = (diag->VM*Pdiag + left->VM*Pleft + top->VM*Ptop)/Psum ;
	current->VC = (diag->VC*Pdiag + left->VC*Pleft + top->VC*Ptop)/Psum ;
	current->VH = (diag->VH*Pdiag + left->VH*Pleft + top->VH*Ptop)/Psum ;
	current->VV = (diag->VV*Pdiag + left->VV*Pleft + top->VV*Ptop)/Psum ;
}

/*
 *	compute the current cell from the three adjacent incomming cell
 *	this routine depends on the global variables costDH, costHH, ... etc
 */
ComputeCurrentCell(current,hori,ddiag,vert,match)
	TScell *current, *hori, *ddiag, *vert ;
	bool match ; /* true if the current characters matches */
{
	float CDD, CHD, CVD, CDH, CHH, CVH, CDV, CHV, CVV ;
	TScell left,diag,top ;

	diag = *ddiag ;
	left = *hori ;
	top = *vert ;

	/* the costs of entrance from the diagonal direction */
	if (match) {
		CDD = diag.D+costDM ;
		CHD = diag.H+costHM ;
		CVD = diag.V+costVM ;
	} else {
		CDD = diag.D+costDC ;
		CHD = diag.H+costHC ;
		CVD = diag.V+costVC ;
	}
	/* from the left */
	CDH = left.D+costDH ;
	CHH = left.H+costHH ;
	CVH = left.V+costVH ;
	/* and from the right */
	CDV = top.D+costDV ;
	CHV = top.H+costHV ;
	CVV = top.V+costVV ;

	if (match) TS_weightedincrement(-CDD,-CHD,-CVD,&diag.DM,&diag.HM,&diag.VM) ;
	else       TS_weightedincrement(-CDD,-CHD,-CVD,&diag.DC,&diag.HC,&diag.VC) ;
			
	TS_weightedincrement(-CDH,-CHH,-CVH,&left.DH,&left.HH,&left.VH) ;
	TS_weightedincrement(-CDV,-CHV,-CVV,&top.DV,&top.HV,&top.VV) ;

	current->H = sumlg3(CDH, CHH, CVH) ;
	current->V = sumlg3(CDV, CHV, CVV) ;
	current->D = sumlg3(CDD, CHD, CVD) ;

	TS_weightedaverage(current,&left,&diag,&top) ;
}

#define K 4
/*
 *	compute the various cost from adaptive coding
 */
static TS_AdaptiveCost(left,diag,top,equalchar)
	TScell *left,*diag,*top ;
	bool equalchar ;
{
	float PDM,PHM,PVM,PDC,PHC,PVC,PDH,PVH,PHH,PDV,PHV,PVV ;
	float log12 ;

	log12 = (float) log2((double)12) ;
	if (equalchar) {
		PDM = (diag->DM+1)/(sumD(diag)+K) ;
		PHM = (diag->HM+1)/(sumH(diag)+K) ; 
		PVM = (diag->VM+1)/(sumV(diag)+K) ; 
		costDM  = (float)log2((double)PDM) - 2 ;
		costHM  = (float)log2((double)PHM) - 2 ;
		costVM  = (float)log2((double)PVM) - 2 ;
	} else {
		PDC = (diag->DC+1)/(sumD(diag)+K) ;
		PHC = (diag->HC+1)/(sumH(diag)+K) ; 
		PVC = (diag->VC+1)/(sumV(diag)+K) ; 
		costDC  = (float)log2((double)PDC) - log12 ;
		costHC  = (float)log2((double)PHC) - log12 ;
		costVC  = (float)log2((double)PVC) - log12 ;
	}

	PDH = (left->DH+1)/(sumD(left)+K) ; 
	PVH = (left->VH+1)/(sumV(left)+K) ; 
	PHH = (left->HH+1)/(sumH(left)+K) ; 
	costDH  = (float)log2((double)PDH) - 2 ;
	costHH  = (float)log2((double)PHH) - 2 ;
	costVH  = (float)log2((double)PVH) - 2 ;

	PDV = (top->DV+1)/(sumD(top)+K) ; 
	PVV = (top->VV+1)/(sumV(top)+K) ; 
	PHV = (top->HV+1)/(sumH(top)+K) ; 
	costDV  = (float)log2((double)PDV) - 2 ;
	costHV  = (float)log2((double)PHV) - 2 ;
	costVV  = (float)log2((double)PVV) - 2 ;
}


/*
 *	The last argument is the forward matrix.
 *	The calling routine can choose to supply NULL to this argument
 *	when the matrix is not required.
 */
static DPA( mode, P, result, matrix)
	int  mode ;
	TScell *result ;
	ThreeStatePBlock *P ;
	TSpcell matrix[][MATRIXSIZE] ;
{
	TScell c1[MAXSEQLENGTH], c2[MAXSEQLENGTH] ;
	TScell *d1, *d2, *temp, boundarycell ;
	int i, j ;
	float log12 ;

	TS_initcell(&boundarycell) ;
	log12 = (float) log2((double)(12)) ;
	if (mode == fixed)
		TS_ProbToCosts(P,&costDM,&costDC,&costDH,&costDV,&costHM,&costHC,&costHH,&costHV,&costVM,&costVC,&costVH,&costVV) ;

	d1 = c1 ;
	d2 = c2 ;
	/* Boundary Conditions */
	TS_initcell(&d1[0]) ;
	d1[0].D = (float)log2((double)P->D) ;
	d1[0].H = (float)log2((double)P->H) ;
	d1[0].V = (float)log2((double)P->V) ;
	if (matrix) copy(&matrix[0][0],&d1[0]) ;
	debug(TS_printcell(0,0,&d1[0]) ;)

	for (j = 1 ; j <= n; j++) {
		if (mode == adaptive)
			TS_AdaptiveCost(&d1[j-1],&boundarycell,&boundarycell,FALSE) ;
		ComputeCurrentCell(&d1[j],&d1[j-1],&boundarycell,&boundarycell,FALSE) ;
		if (matrix != NULL) copy(&matrix[0][j],&d1[j]) ;
		debug(TS_printcell(0,j,&d1[j]) ;)
	}
 
	for (i = 1 ; i <= m ; i++) { /* General Step */
		if (mode == adaptive)
			TS_AdaptiveCost(&boundarycell,&boundarycell,&d1[0],FALSE) ;
		ComputeCurrentCell(&d2[0],&boundarycell,&boundarycell,&d1[0],FALSE) ;
		if (matrix != NULL) copy(&matrix[i][0],&d2[0]) ;
		debug(TS_printcell(i,0,&d2[0]) ;)

		for (j = 1; j <= n; j++) {
			if (mode == adaptive )
				TS_AdaptiveCost(&d2[j-1],&d1[j-1],&d1[j],a[i]==b[j]) ;
			ComputeCurrentCell(&d2[j],&d2[j-1],&d1[j-1],&d1[j],a[i]==b[j]) ;
			if (matrix != NULL) copy(&matrix[i][j],&d2[j]) ;
			debug(TS_printcell(i,j,&d2[j]) ;)
		} /* for j */
		swap(d1, d2) ;
	} /* for i */
	*result = d1[n] ;
} /*DPA*/

/*
 *	Fwd[i][j] given by DPA gives the prob that the machine ends in the
 *	various states after generating A[1..i], B[1..j].
 *	This routine computes the matrix Rev[i][j] which gives the prob
 *	that FROM the various starting state the machine will
 *	generate A[i+1..m], B[j+1..n].
 *	So Fwd[i][j] + Rev[i][j] gives the density plot.
 */
static DPArev(P,matrix)
	ThreeStatePBlock *P ;
	TSpcell matrix[][MATRIXSIZE] ;
{
	TSpcell e1[MAXSEQLENGTH], e2[MAXSEQLENGTH] ;
	TSpcell *d1, *d2, *temp ;
	int i, j ;

	TS_ProbToCosts(P,&costDM,&costDC,&costDH,&costDV,&costHM,&costHC,&costHH,&costHV,&costVM,&costVC,&costVH,&costVV) ;

	d1 = e1 ;
	d2 = e2 ;
	/* Boundary Conditions */
	d1[n+1].D = d1[n+1].H = d1[n+1].V = 0 ;
	matrix[m+1][n+1] = d1[n+1] ;
      
	for (j = n ; j >= 1; j--) {
		d1[j].D = d1[j+1].H + costDH ;
		d1[j].H = d1[j+1].H + costHH ;
		d1[j].V = d1[j+1].H + costVH ;
		matrix[m+1][j] = d1[j] ;
	}
 
	for (i = m ; i >= 1 ; i--) { /* General Step */
		d2[n+1].D = d1[n+1].V + costDV ;
		d2[n+1].H = d1[n+1].V + costHV ;
		d2[n+1].V = d1[n+1].V + costVV ;
		matrix[i][n+1] = d2[n+1] ;

		for (j = n; j >= 1; j--) {
			d2[j].D = sumlg3(d2[j+1].H + costDH,
					d1[j+1].D + (a[i]==b[j]?costDM:costDC),
					d1[j].V + costDV) ;
			d2[j].H = sumlg3(d2[j+1].H + costHH,
					d1[j+1].D + (a[i]==b[j]?costHM:costHC),
					d1[j].V + costHV) ;
			d2[j].V = sumlg3(d2[j+1].H + costVH,
					d1[j+1].D + (a[i]==b[j]?costVM:costVC),
					d1[j].V + costVV) ;
			matrix[i][j] = d2[j] ;
		} /* for j */
		swap(d1, d2) ;
	} /* for i */
} /*DPArev*/

/*--------------------------------------------------------------------------
 *	Computes the 3 state R-theory message length with a fixed set of parameters
 */
FixParmThreeState(A,B,P,result)
	string *A, *B ;
	ThreeStatePBlock *P ;
	TScell *result ;
{
	a = A->sequence ; b = B->sequence ;
	m = A->length ; n = B->length ;
	DPA(fixed,P,result,(TSpcell *) NULL) ;
}


/*--------------------------------------------------------------------------*/

#define D 0 
#define H 1 
#define V 2 
#define M 3 
#define C 4 

static int
transition(PM, PC, PH)
	float	PM, PC, PH ;
{
	float t ;
	int	event ;

	t = (float)(rand() % 10000000) / 10000000 ; /* 0 <= t < 1 */
	if ( (t -= PM) < 0) event = M ;
	else if ((t -= PC) < 0) event = C ;
	else if ((t -= PH) < 0) event = H ;
	else event = V ;

	return event ;
}

/*
 *	A 3 State machine to generate a pair of strings according to
 *	the given set of parameters.
 *	Length refer to the average length of the two strings
 *
 *	If Alignment is not NULL then it contains the generative
 *	alignment of the strings.
 */
ThreeStateGen(A, B, length, P, Alignment, Alignlength)
	string *A, *B ;
	ThreeStatePBlock *P ;
	align *Alignment ;
	int *Alignlength ;
{
	int state, event, i, alignlen ;
	int	alen, blen ;
	symbol *s ;
	symbol a[MAXSEQLENGTH], b[MAXSEQLENGTH] ;

	alen = blen = 0 ;
	state = D ;
	alignlen = 0 ;
	while ((alen + blen) < 2*length) {
		switch ( state ) {
		case D :
			event = transition(P->DM, P->DC, P->DH) ;
			break ;
		case H :
			event = transition(P->HM, P->HC, P->HH) ;
			break ;
		case V :
			event = transition(P->VM, P->VC, P->VH) ;
			break ;
		}
		if (event == M) {
			state = D ;
			a[++alen] = b[++blen] = randomsymbol(&DNASymbol) ;
			if (Alignment) {
				Alignment[++alignlen].a = alen ;
				Alignment[alignlen].b = blen ;
			}
		} else if (event == C) {
			state = D ;
			randomsymbolpair(&DNASymbol, &a[++alen], &b[++blen] ) ;
			if (Alignment) {
				Alignment[++alignlen].a = alen ;
				Alignment[alignlen].b = blen ;
			}
		} else if (event == H) {
			state = H ;
			a[++alen] = randomsymbol(&DNASymbol) ;
			if (Alignment) {
				Alignment[++alignlen].a = alen ;
				Alignment[alignlen].b = 0 ;
			}
		} else {
			state = V ;
			b[++blen] = randomsymbol(&DNASymbol) ;
			if (Alignment) {
				Alignment[++alignlen].a = 0 ;
				Alignment[alignlen].b = blen ;
			}
		}

		if (alen > MAXSEQLENGTH || blen > MAXSEQLENGTH) {
			printf("ThreeStateGen : string too long\n") ;
			exit(99) ;
		}
	}
	s = A->sequence = (symbol *) malloc((unsigned)(alen+1)*sizeof(symbol)) ;
	A->length = alen ;
	s[0] = NULLSymbol ;
	for (i = 1; i <= alen; i++) s[i] = a[i] ;

	s = B->sequence = (symbol *) malloc((unsigned)(blen+1)*sizeof(symbol)) ;
	B->length = blen ;
	s[0] = NULLSymbol ;
	for (i = 1; i <= blen; i++) s[i] = b[i] ;

	if (Alignment) *Alignlength = alignlen ;
}

