
/*
 *  Copyright  (c)  L.Allison, C.S.Wallace, C.N.Yee.
 *     Dept. Computer Science, Monash University, Australia 3168,    May 1991.
 *  1. These routines may only be used for non-commercial, non-classified,
 *     research purposes.
 *  2. No liability of any kind is accepted for their use.
 *  3. As a courtesy, please inform us of any interesting uses of the routines,
 *     of any bugs that you find in them and of how to correct
 *     such bugs, if known.
 *  4. Do not delete this copyright notice.
 *  5. Please include the following references in any publication based wholly
 *     or in part on the use of these routines or on the ideas contained in them:
 *   . Allison L. & C.N.Yee.
 *        Minimum Message Length Encoding and the Comparison of Macro-Molecules.
 *        Bull. Math. Bio. 52(3) 431-453 1990
 *   . Allison L., C.S.Wallace & C.N.Yee. Finite-State Models in the Alignment
 *        of Macromolecules. J. Mol. Evol. (1992) 35:77-89
 *  6. You may distribute the routines to others if and only if
 *     they agree to accept and maintain these conditions and no fee is charged.
 *  7. There is no rule 7.
 */
/*	"3StateR0.c"  best alignment under the 3State model
 *
 *	Compute the best alignment of 2 strings under 3 state R-theory
 *	Requires some routines in 3State.c
 *
 *	Entries:
 *
 * ThreeStateR0ML(A,B,MesgLength,MatrixPart,P,Symmetric)
 *	string *A, *B ;
 *	int Symmetric ; - if true then force state H and V to have same parameters
 *	float *MesgLength,*MatrixPart ; - the message length, and the matrix part of the ML
 *
 */
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "macro.h"
#include "Strings.h"
#include "rtheory.h"

#define MAXITERATION	40

/*
 *	some global variables
 */
static symbol   *a, *b ;
static int m,n ;

static float costDM, costDC, costDH, costDV ;
static float costHM, costHC, costHH, costHV ;
static float costVM, costVC, costVH, costVV ;

#define Horiz 0
#define Vert 1
#define Diag 2

static TSR0_DPA(P,Alignment,Alignlength,MatrixPart)
	ThreeStatePBlock *P ;
	align Alignment[] ;
	int *Alignlength ;
	float *MatrixPart ;
{
	char wherefrom[MATRIXSIZE][MATRIXSIZE] ;
	TSpcell c1[MAXSEQLENGTH], c2[MAXSEQLENGTH] ;
	TSpcell *d1, *d2, *d, *temp ;
	int i, j ;
	int x, y, tempa, tempb, mid ;

	TS_ProbToCosts(P,&costDM,&costDC,&costDH,&costDV,&costHM,&costHC,&costHH,&costHV,&costVM,&costVC,&costVH,&costVV) ;

	d1 = c1 ;
	d2 = c2 ;
	/* Boundary Conditions */
	d1[0].D = d1[0].H = d1[0].V = 0 ;

	for (j = 1 ; j <= n; j++) {
		d1[j].H = d1[j-1].H + costHH ;
		d1[j].D = d1[j].V = -infinity ;
		wherefrom[0][j] = Horiz ;
	}
 
	for (i = 1 ; i <= m ; i++) { /* General Step */
		d2[0].V = d1[0].V + costVV ;
		d2[0].D = d2[0].H = -infinity ;
		wherefrom[i][0] = Vert ;
		for (j = 1; j <= n; j++) {

			d = &d1[j-1] ;
			if (a[i] == b[j])
				d2[j].D =fmax3(d->D+costDM,d->V+costVM,d->H+costHM);
			else
				d2[j].D = fmax3(d->D+costDC,d->V+costVC,d->H+costHC) ;

			d = &d2[j-1] ;
			d2[j].H = fmax3(d->D+costDH,d->H+costHH,d->V+costVH) ;

			d = &d1[j] ;
			d2[j].V = fmax3(d->D+costDV,d->H+costHV,d->V+costVV) ;

			if (d2[j].D >= d2[j].H && d2[j].D >= d2[j].V)
				wherefrom[i][j] = Diag ;
			else if (d2[j].H >= d2[j].D && d2[j].H >= d2[j].V)
				wherefrom[i][j] = Horiz ;
			else
				wherefrom[i][j] = Vert ;
		} /* for j */
		swap(d1, d2) ;
	} /* for i */
	*MatrixPart = -fmax3(d1[n].D, d1[n].H, d1[n].V) ;

	/*
	 *	trace alignment
	 */
	x = m; y = n ;
	*Alignlength = 0 ;
	while (x != 0 || y != 0) {
		(*Alignlength) ++ ;
		if (wherefrom[x][y] == Diag)  {
			Alignment[*Alignlength].a = x ;
			Alignment[*Alignlength].b = y ;
			x--; y-- ;
		} else if (wherefrom[x][y] == Horiz) {
			Alignment[*Alignlength].a = 0 ;
			Alignment[*Alignlength].b = y ;
			y-- ;
		} else {
			Alignment[*Alignlength].a = x ;
			Alignment[*Alignlength].b = 0 ;
			x-- ;
		}
	}
	/* reverse direction */
	mid = *Alignlength / 2 ;
	for (i = 1, j= *Alignlength ; i <= mid; i++, j--) {
		tempa = Alignment[i].a ;
		tempb = Alignment[i].b ;
		Alignment[i].a = Alignment[j].a ;
		Alignment[i].b = Alignment[j].b ;
		Alignment[j].a = tempa ;
		Alignment[j].b = tempb ;
	}

} /*TSR0_DPA*/

TSR0_ComputeParameters(Alignment,Alignlength,A,B,P,MesgLength)
	align Alignment[] ;
	int Alignlength ;
	symbol A[], B[] ;
	ThreeStatePBlock *P ;
	float *MesgLength ;
{
	int i,laststate ;
	float DM,DC,DH,DV,HM,HC,HH,HV,VM,VC,VH,VV ;
	float ML ;
	float log12 ;

	log12 = log2((double)12) ;
	DM=DC=DH=DV=HM=HC=HH=HV=VM=VC=VH=VV=0 ;
	ML = 0 ;
	laststate = Diag ;
	for (i=1; i <= Alignlength ; i++) {
		if (Alignment[i].a != 0 && Alignment[i].b != 0) { /* diag move */
			switch (laststate) {
				case Diag :
					if (A[Alignment[i].a] == B[Alignment[i].b]) {
						DM++ ;
						ML += -log2((double) DM/(DM+DC+DH+DV+3)) +2 ;
					} else {
						DC ++ ;
						ML += -log2((double) DC/(DM+DC+DH+DV+3)) + log12 ;
					}
					break ;
				case Horiz :
					if (A[Alignment[i].a] == B[Alignment[i].b]) {
						HM ++ ;
						ML += -log2((double) HM/(HM+HC+HH+HV+3)) +2 ;
					} else {
						HC ++ ;
						ML += -log2((double) HC/(HM+HC+HH+HV+3)) + log12 ;
					}
					break ;
				case Vert :
					if (A[Alignment[i].a] == B[Alignment[i].b]) {
						VM ++ ;
						ML += -log2((double) VM/(VM+VC+VH+VV+3)) +2 ;
					} else {
						VC ++ ;
						ML += -log2((double) VC/(VM+VC+VH+VV+3)) + log12 ;
					}
					break ;
			}
			laststate = Diag ;
		} else if (Alignment[i].a == 0 ) { /* Horiz move */
			switch (laststate) {
				case Diag :
					DH++ ;
					ML += -log2((double) DH/(DM+DC+DH+DV+3)) +2 ;
					break ;
				case Horiz :
					HH++ ;
					ML += -log2((double) HH/(HM+HC+HH+HV+3)) +2 ;
					break ;
				case Vert :
					VH++ ;
					ML += -log2((double) VH/(VM+VC+VH+VV+3)) +2 ;
					break ;
			}
			laststate = Horiz ;
		} else {
			switch (laststate) {
				case Diag :
					DV++ ;
					ML += -log2((double) DV/(DM+DC+DH+DV+3)) +2 ;
					break ;
				case Horiz :
					HV++ ;
					ML += -log2((double) HV/(HM+HC+HH+HV+3)) +2 ;
					break ;
				case Vert :
					VV++ ;
					ML += -log2((double) VV/(VM+VC+VH+VV+3)) +2 ;
					break ;
			}
			laststate = Vert ;
		}
	}
	DH=DV=(DH+DV)/2 ;
	HM=VM=(HM+VM)/2 ;
	HC=VC=(HC+VC)/2 ;
	HH=VV=(HH+VV)/2 ;
	HV=VH=(HV+VH)/2 ;
	fchangetoprobability4(DM,DC,DH,DV,&P->DM,&P->DC,&P->DH,&P->DV) ;
	fchangetoprobability4(HM,HC,HH,HV,&P->HM,&P->HC,&P->HH,&P->HV) ;
	fchangetoprobability4(VM,VC,VH,VV,&P->VM,&P->VC,&P->VH,&P->VV) ;
	*MesgLength = ML ;
} /* TSR0_ComputeParameters() */


/*
 *	look through the stack to detect cycle
 */
bool found(val, stack, maxstack)
	float val, stack[] ;
	int maxstack ;
{
	int i ;

	for (i=0; i< maxstack; i++)
		if (stack[i] == val) return TRUE ;
	return FALSE ;
}
/*--------------------------------------------------------------------------
 *	Computes the 3 states R-theory message length, and return the optimal probs
 */
ThreeStateR0ML(A,B,Alignment,Alignlength,MesgLength,P)
	string *A, *B ;
	float *MesgLength ;
	align Alignment[] ;
	int *Alignlength ;
	ThreeStatePBlock *P ;
{
	float MatrixPart ;
	int iterations ;
	float stack[MAXITERATION] ;	 /* for cycle detection */

	a = A->sequence ; b = B->sequence ;
	m = A->length ;   n = B->length ;

	iterations = 0 ;

	printf("  Iteration  0 :\n") ;
	*MesgLength = infinity ;
	do {
		TS_printparameters(P) ;
		TSR0_DPA(P,Alignment,Alignlength,&MatrixPart) ;
		TSR0_ComputeParameters(Alignment,*Alignlength,a,b,P,MesgLength) ;
		printf("    Matrix Part = %7.2f  ML = %7.2f\n",MatrixPart,*MesgLength) ;

		fflush(stdout) ;
		stack[iterations] = *MesgLength ;
		iterations++ ;
		if (found(*MesgLength,stack,iterations-1)) {/* cycle deteted */
			if (*MesgLength != stack[iterations-2])
				printf("....CYCLE DETECTED....\n") ;
			break ;
		}
		if (iterations > MAXITERATION) {
			printf("3State R-0 : max iteration exceeded\n") ;
			break ;
		}
		printf("  Iteration %2d :\n",iterations) ;
	} while (TRUE) ;
}
