/*
 *  Copyright  (c)  L.Allison, C.S.Wallace, C.N.Yee.
 *     Dept. Computer Science, Monash University, Australia 3168,    May 1991.
 *  1. These routines may only be used for non-commercial, non-classified,
 *     research purposes.
 *  2. No liability of any kind is accepted for their use.
 *  3. As a courtesy, please inform us of any interesting uses of the routines,
 *     of any bugs that you find in them and of how to correct
 *     such bugs, if known.
 *  4. Do not delete this copyright notice.
 *  5. Please include the following references in any publication based wholly
 *     or in part on the use of these routines or on the ideas contained in them:
 *   . Allison L. & C.N.Yee.
 *        Minimum Message Length Encoding and the Comparison of Macro-Molecules.
 *        Bull. Math. Bio. 52(3) 431-453 1990
 *   . Allison L., C.S.Wallace & C.N.Yee. Finite-State Models in the Alignment
 *        of Macromolecules. J. Mol. Evol. (1992) 35:77-89
 *  6. You may distribute the routines to others if and only if
 *     they agree to accept and maintain these conditions and no fee is charged.
 *  7. There is no rule 7.
 */
/*	"5State.c"  prefix FS
 *
 *	5 States machine ...  extension of the fully connected
 *	3 States model with two 'ears' for long indels
 *	
 */
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "macro.h"
#include "Strings.h"
#include "rtheory.h"

#define FREEADAPT	/* let everything adapt freely on 1st adaptive run */
#undef FREEADAPT

#define ProbAA	0.99	/* the fixed probs for the long indels */
#define ProbAD	0.01
#define ProbDA	0.001

/* forward declarations */
float sumD(), sumH(), sumV(), sumA(), sumB() ;

/*
 *	some global variables
 */
static symbol   *a, *b ;
static int m,n ;

static float costDM, costDC, costDH, costDV, costDA, costDB ;
static float costHM, costHC, costHH, costHV ;
static float costVM, costVC, costVH, costVV ;
static float costAD, costAA ;
static float costBD, costBB ;

/*---------------------------------------------------------------------------
 *	Computes the corresponding edit cost from a set of arcs probabilities
 */
FS_ProbToCosts(P)
	FiveStatePBlock *P ;
{
	float log12 ;

	log12 = (float) log2((double)(12)) ;

	costDM  = (float)log2((double)P->DM) - 2 ;
	costDC  = (float)log2((double)P->DC) - log12 ;
	costDH  = (float)log2((double)P->DH) - 2 ;
	costDV  = (float)log2((double)P->DV) - 2 ;
	costDA  = (float)log2((double)P->DA) - 2 ;
	costDB  = (float)log2((double)P->DB) - 2 ;

	costHM  = (float)log2((double)P->HM) - 2 ;
	costHC  = (float)log2((double)P->HC) - log12 ;
	costHH  = (float)log2((double)P->HH) - 2 ;
	costHV  = (float)log2((double)P->HV) - 2 ;

	costVM  = (float)log2((double)P->VM) - 2 ;
	costVC  = (float)log2((double)P->VC) - log12 ;
	costVH  = (float)log2((double)P->VH) - 2 ;
	costVV  = (float)log2((double)P->VV) - 2 ;

	costAD  = (float)log2((double)P->AD) - 2 ;
	costAA  = (float)log2((double)P->AA) - 2 ;

	costBD  = (float)log2((double)P->BD) - 2 ;
	costBB  = (float)log2((double)P->BB) - 2 ;
}

/*------------------------------------------------------------------------*/
static float MessageLength(C)
	FScell *C ;
{ 
	float	t1,t2,t3,t4 ;

	t1 = paramcost6(C->DM, C->DC, C->DH, C->DV, C->DA, C->DB) ;
	t2 = paramcost4(C->HM, C->HC, C->HH, C->HV) + paramcost4(C->VM, C->VC, C->VV, C->VH) + paramcost2(C->AD, C->AA) + paramcost2(C->BD, C->BB) ;
	t3 = -sumlg5(C->D, C->H, C->V, C->A, C->B) ;
	t4 =  U(sumH(C) + sumV(C) + sumD(C) + sumA(C) + sumB(C)) ;
	return (t1+t2+t3+t4) ;
}

/*
 *	Message length when the probs for the long indels are fixed,
 *	so need not be coded
 */
static float MessageLength2(C)
	FScell *C ;
{ 
	float	t1,t2,t3 ;

	/* parameters cost */
	t1 = paramcost4(C->DM, C->DC, C->DH, C->DV) +
 	     paramcost4(C->HM, C->HC, C->HH, C->HV) +
		 paramcost4(C->VM, C->VC, C->VV, C->VH) ;
	/* the matrix part */
	t2 = -sumlg5(C->D, C->H, C->V, C->A, C->B) ;
	/* number of instructions */
	t3 =  U(sumH(C) + sumV(C) + sumD(C) + sumA(C) + sumB(C)) ;
	return (t1+t2+t3) ;
}

/*--------------------------------------------------------------------------
 *	Some general utility routines
 */
static float sumD(C) FScell *C ; { return C->DM + C->DC + C->DH + C->DV + C->DA + C->DB ; }
static float sumH(C) FScell *C ; { return C->HM + C->HC + C->HH + C->HV ; }
static float sumV(C) FScell *C ; { return C->VM + C->VC + C->VH + C->VV ; }
static float sumA(C) FScell *C ; { return C->AD + C->AA ; }
static float sumB(C) FScell *C ; { return C->BD + C->BB ; }

static copy(p, q) FScell *q; FSpcell *p ;
	{ p->D=q->D; p->H=q->H; p->V=q->V; p->A=q->A; p->B=q->B;  }

static FS_initcell(p)
	FScell *p ;
{
	p->D = p->H =  p->V = p->A = p->B = BIG ;
	p->DM = p->DC = p->DH = p->DV = p->DA = p->DB = 0.0 ;
	p->HM = p->HC = p->HH = p->HV = 0.0 ;
	p->VM = p->VC = p->VH = p->VV = 0.0 ;
	p->AD = p->AA = 0.0 ;
	p->BD = p->BB = 0.0 ;
}

static FS_initpcell(p)
	FSpcell *p ;
{ p->D = p->H = p->V = p->A = p->B = BIG ; }

/*-------------------------------------------------------------------------*/

#ifdef debug
TS_printcell(i,j,c)
	int i,j;
	FScell *c ;
{
	printf("(%2d,%2d) %4.2f %4.2f %4.2f %4.2f %4.2f",
		i,j,c->D,c->H,c->V,c->A,c->B);
	printf(" | %4.2f %4.2f %4.2f %4.2f %4.2f %4.2f\n",
		c->DM,c->DC,c->DH,c->DV,c->DA,c->DB) ;
	printf("       (H) %4.2f %4.2f %4.2f %4.2f (V) %4.2f %4.2f %4.2f %4.2f (A) %4.2f %4.2f (B) %4.2f %4.2f\n",
		c->HM,c->HC,c->HH,c->HV,c->VM,c->VC,c->VV,c->VH,c->AD,c->AA, c->BD, c->BB) ;
}
#endif

FS_printparameters(P)
	FiveStatePBlock *P ;
{
	printf("    D %5.3f(M) %5.3f(C) %5.3f(H) %5.3f(V) %5.3f(A) %5.3f(B)\n",P->DM,P->DC,P->DH,P->DV,P->DA,P->DB) ;
	printf("    H %5.3f(M) %5.3f(C) %5.3f(H) %5.3f(V)\n",P->HM,P->HC,P->HH,P->HV) ;
	printf("    V %5.3f(M) %5.3f(C) %5.3f(H) %5.3f(V)\n",P->VM,P->VC,P->VH,P->VV) ;
	printf("    A %5.3f(D) %5.3f(A)\n",P->AD,P->AA) ;
	printf("    B %5.3f(D) %5.3f(B)\n",P->BD,P->BB) ;
}

FS_printparameters2(P,result,MesgLength)
	FiveStatePBlock *P ;
	float MesgLength ;
	FScell *result ;
{
	printf("    D %5.4f %5.3f(M) %5.3f(C) %5.3f(H) %5.3f(V) %5.3f(A) %5.3f(B) %5.1f(sumfq)\n",P->D,P->DM,P->DC,P->DH,P->DV,P->DA,P->DB,sumD(result)) ;
	printf("    H %5.4f %5.3f(M) %5.3f(C) %5.3f(H) %5.3f(V) %5.1f(sumfq)\n",P->H,P->HM,P->HC,P->HH,P->HV, sumH(result)) ;
	printf("    V %5.4f %5.3f(M) %5.3f(C) %5.3f(H) %5.3f(V) %5.1f(sumfq)\n",P->V,P->VM,P->VC,P->VH,P->VV, sumV(result)) ;
	printf("    A %5.4f %5.3f(D) %5.3f(A) %5.1f(sumfq)\n",P->A,P->AD,P->AA,sumA(result)) ;
	printf("    B %5.4f %5.3f(D) %5.3f(B) %5.1f(sumfq)\n",P->B,P->BD,P->BB,sumB(result)) ;
	printf("    ML=%10.4f (matrix part=%10.4f)\n", MesgLength,-sumlg5(result->D,result->H,result->V,result->A,result->B)) ;

}

FS_statistics(P,result,MesgLength)
	FiveStatePBlock *P ;
	float *MesgLength ;
	FScell *result ;
{
	float sum ;

	fchangetoprobability6(result->DM,result->DC,result->DH,result->DV,result->DA
		,result->DB,&P->DM,&P->DC,&P->DH,&P->DV,&P->DA,&P->DB) ;
	fchangetoprobability4(result->HM,result->HC,result->HH,result->HV,&P->HM,&P->HC,&P->HH,&P->HV) ;
	fchangetoprobability4(result->VM,result->VC,result->VH,result->VV,&P->VM,&P->VC,&P->VH,&P->VV) ;
	fchangetoprobability2(result->AD,result->AA,&P->AD,&P->AA) ;
	fchangetoprobability2(result->BD,result->BB,&P->BD,&P->BB) ;

	/* prob that the machine is in the various states as determined by
	   the weighted count */
	sum = sumD(result)+sumH(result)+sumV(result)+sumA(result)+sumB(result);
	P->D = (result->DM+result->DC+result->HM+result->HC+result->VM+result->VC+result->AD+result->BD+0.5)/(sum+2.5) ;
	P->H = (result->DH+result->HH+result->VH+0.5)/(sum+2.5) ;
	P->V = (result->DV+result->VV+result->HV+0.5)/(sum+2.5) ;
	P->A = (result->DA+result->AA+0.5)/(sum+2.5) ;
	P->B = (result->DB+result->BB+0.5)/(sum+2.5) ;

	*MesgLength = MessageLength(result) ;
}

/*
 *	fix the probs for the long indels
 */
FS_statistics2(P,result,MesgLength)
	FiveStatePBlock *P ;
	float *MesgLength ;
	FScell *result ;
{
	float sum, f ;

	fchangetoprobability4(result->DM,result->DC,result->DH,result->DV,
		&P->DM,&P->DC,&P->DH,&P->DV) ;
	fchangetoprobability4(result->HM,result->HC,result->HH,result->HV,&P->HM,&P->HC,&P->HH,&P->HV) ;
	fchangetoprobability4(result->VM,result->VC,result->VH,result->VV,&P->VM,&P->VC,&P->VH,&P->VV) ;

	P->DA = P->DB = ProbDA ;
	P->AA = P->BB = ProbAA ;
	P->AD = P->BD = ProbAD ;

	/*	P->DM+P->DC+P->DH+P->DV equals 1 now
	 *	make adjustments so that P->DM+P->DC+P->DH+P->DV+P->DA+P->DB equals 1 
	 */
	f = 1 - 2*ProbDA ;
	P->DM *= f ;
	P->DC *= f ;
	P->DH *= f ;
	P->DV *= f ;

	/* prob that the machine is in the various states as determined by
	   the weighted count */
	sum = sumD(result)+sumH(result)+sumV(result)+sumA(result)+sumB(result);
	P->D = (result->DM+result->DC+result->HM+result->HC+result->VM+result->VC+result->AD+result->BD+0.5)/(sum+2.5) ;
	P->H = (result->DH+result->HH+result->VH+0.5)/(sum+2.5) ;
	P->V = (result->DV+result->VV+result->HV+0.5)/(sum+2.5) ;
	P->A = (result->DA+result->AA+0.5)/(sum+2.5) ;
	P->B = (result->DB+result->BB+0.5)/(sum+2.5) ;

	*MesgLength = MessageLength2(result) ;
}


/*------------------------------------------------------------------------*/
static printmatrix(M)
	float M[][MATRIXSIZE] ;
{
	int i, j ;

	printf("       ") ;
	for ( j = 1 ; j <= n ; j++) printf("  %c ",SymbTable->Symbol[b[j]]) ;
	putchar('\n') ;
	for ( i = 0 ; i <= m ; i++) {
		if ( i != 0) printf(" %c ",SymbTable->Symbol[a[i]]) ;
		else printf("   ") ;
		for ( j = 0 ; j <= n ; j++)
			/* printf(" %3d",(int) (M[i][j]+0.5)) ;*/
			printf(" %5.0f", M[i][j]) ;
		putchar('\n') ;
	}
}

/*
 *	combine Fwd and Rev
 */
static combine(Fwd,Rev,DMatrix)
	FSpcell Fwd[][MATRIXSIZE],Rev[][MATRIXSIZE] ;
	float DMatrix[][MATRIXSIZE] ;
{
	int i, j ;

	for ( i = 0 ; i <= m ; i++)
		for ( j = 0 ; j <= n ; j++) {
			DMatrix[i][j] = -sumlg5(Fwd[i][j].D+Rev[i+1][j+1].D,
			 		Fwd[i][j].H+Rev[i+1][j+1].H,
			 		Fwd[i][j].V+Rev[i+1][j+1].V,
			 		Fwd[i][j].A+Rev[i+1][j+1].A,
			 		Fwd[i][j].B+Rev[i+1][j+1].B);
		}
}

/*
 *	Do a forward run then a reverse run using the parameters P
 *	then combine the fwd and rev result output into OFILe the
 *	resulting density plot
 */
static DensityPlot(A,B,P,ofile)
	string *A, *B ;
	FiveStatePBlock *P ;
	FILE *ofile ;
{
	FSpcell Fwd[MATRIXSIZE][MATRIXSIZE], Rev[MATRIXSIZE][MATRIXSIZE] ;
	float	DMatrix[MATRIXSIZE][MATRIXSIZE] ;
	int i, j ;
	FScell result ;

	if (ofile || (m < SCREENHEIGHT && n < SCREENWIDTH)) {
		DPA(fixed,P,&result, Fwd) ;
		DPArev(P,Rev) ;

		combine(Fwd,Rev,DMatrix) ;
		scale(DMatrix,m,n);
		if (m < SCREENHEIGHT && n < SCREENWIDTH) PlotMatrix(a,b,DMatrix,m,n,SymbTable) ;

		if (ofile) OutputMatrix(A,B,DMatrix,ofile,"FiveState R-Theory",SymbTable) ;
	}
}
/*--------------------------------------------------------------------------
 *	Computes the 5 states R-theory message length, and return the optimal probs
 */
FiveStateMesgLength(A,B,MesgLength,MatrixPart,mode,outfile,P,Accurarcy,flag)
	string *A, *B ;
	FILE *outfile ;	/* if not NULL the density plot will be generated and
						output to it */
	int mode ;	/* if equal fixeed then use the supplied parameters */
				/* else use adaptive coding for the first run to guess a set
				   of good parameters */
	int flag ;  /* if flag equals FREE then allow all parms to adapt freely */
				/* otherwise fix the probs for the long indels */
	float *MesgLength, *MatrixPart ;
	float Accurarcy ; /* the stopping accurarcy */
	FiveStatePBlock *P ;
{
	float ML, minML, minMatrixPart ;
	int iterations ;
	FScell result ;

	char filename[100] ;

	a = A->sequence ; b = B->sequence ;
	m = A->length ; n = B->length ;


	if (outfile && (m > MATRIXSIZE || n > MATRIXSIZE)) {
		printf("FiveStateML() : Warning : string too long, no density plot produced\n") ;
		outfile = (FILE *) NULL ;
	}

	 
	*MesgLength = minML = minMatrixPart = infinity ;
	iterations = 1 ;
	if (mode==adaptive) P->D = P->H = P->V = P->A = P->B = 1.0/5 ;

	do {
		ML = *MesgLength ;
		DPA(mode,P,&result,(FSpcell *) NULL) ;

		if (flag == FREE) FS_statistics(P,&result,MesgLength) ;
		else FS_statistics2(P,&result,MesgLength) ;

		*MatrixPart = -sumlg5(result.D,result.H,result.V,result.A,result.B) ;

		printf("Iteration %2d :\n",iterations) ;
		FS_printparameters2(P,&result,*MesgLength) ;
		fflush(stdout) ;

		mode = fixed;
		iterations++ ;
		if (*MesgLength < minML) {
			minML = *MesgLength ;
			minMatrixPart = *MatrixPart ;
		}
	} while (flabs((ML-*MesgLength)/ *MesgLength) > Accurarcy) ;

	printf("minimun MesgLength is %7.2f\n",minML) ;
	*MesgLength = minML ;
	*MatrixPart = minMatrixPart ;

	DensityPlot(A,B,P,outfile) ;
}

/*=======================================================================*/

/*
 *	prob of entering from each of the directions,  scaled up by MLmin
 *  so that at least one of them equal 1
 */
static scaledprobability5(L1, L2, L3, L4, L5, P1, P2, P3, P4, P5, Psum)
	float L1, L2, L3, L4, L5 ;
	float *P1, *P2, *P3, *P4, *P5, *Psum ;
{
	float Lmin ;
	Lmin = fmin5(L1, L2, L3, L4, L5) ;
	*P1 = exp2((double)Lmin-L1) ;
	*P2 = exp2((double)Lmin-L2) ;
	*P3 = exp2((double)Lmin-L3) ;
	*P4 = exp2((double)Lmin-L4) ;
	*P5 = exp2((double)Lmin-L5) ;
	*Psum = *P1+*P2+*P3+*P4+*P5 ;
}

/*
 *	a count of 1 is distrubuted to the three counters S1, S2 and S3 
 *	weighted by the message length (log probabilities) L1, L2 and L3.
 */
static scaledprobability3(L1,L2,L3,P1,P2,P3,Psum)
	float L1,L2,L3,*P1,*P2,*P3,*Psum ;
{
	float Lmin ;

	Lmin = fmin3(L1,L2,L3) ;
	*P1 = exp2((double)Lmin-L1) ;
	*P2 = exp2((double)Lmin-L2) ;
	*P3 = exp2((double)Lmin-L3) ;
	*Psum = *P1 + *P2 + *P3 ;
}

/*
 *	L1 corresponds to the log prob of self transition (HH or VV)
 *	L2 corresponds to the log prob of transition from D to H (or V),
 *  and L3, L4, L5 correspond to the transitions that have to done via state
 *	D (HA and VA).
 */
static weightedincrement2(L1, L2, L3, L4, L5, S1, S2, S3, S4, S5)
	float	L1, L2, L3, L4, L5 ;
	float	*S1, *S2, *S3, *S4, *S5 ;
{
	float P1, P2, P3, P4, P5, Psum ;

	scaledprobability5(L1,L2,L3,L4,L5,&P1,&P2,&P3,&P4,&P5,&Psum) ;
	*S1 += P1/Psum ;
	*S2 += (P2+P3+P4+P5)/Psum ;
	*S3 += P3/Psum ;
	*S4 += P4/Psum ;
	*S5 += P5/Psum ;
}

/*
 *	similar to weightedincrement2
 *	L1 corresponds to the log prob of self transition (AA or BB)
 *	L2 corresponds to the log prob of transition from D to H (or V or A or B),
 *  and L3 corresponds to the transition that have to done via state D
 *	(BA or AB).
 */
static weightedincrement3(L1, L2, L3, S1, S2, S3)
	float	L1, L2, L3 ;
	float	*S1, *S2, *S3 ;
{
	float P1, P2, P3, Psum ;

	scaledprobability3(L1,L2,L3,&P1,&P2,&P3,&Psum) ;
	*S1 += P1/Psum ;
	*S2 += (P2+P3)/Psum ;
	*S3 += P3/Psum ;
}

/*
 *	transition into state D.  If current state is H (or V or A or B)
 *	we have to make a transition from it to state D first
 *	then in all three cases we will make a definite transition of
 *	DM (if we have a match) or DC (if we have a change).
 *	Refer to how this routine is called.
 */
static weightedincrement(L1, L2, L3, L4, L5, S1, S2, S3, S4, S5)
	float	L1, L2, L3, L4, L5 ;
	float	*S1, *S2, *S3, *S4, *S5 ;
{
	float P1, P2, P3, P4, P5, Psum ;

	scaledprobability5(L1,L2,L3,L4,L5,&P1,&P2,&P3,&P4,&P5,&Psum) ;
	*S1 += 1 ;  /* in all three cases we have to increment this counter */
	*S2 += P2/Psum ;
	*S3 += P3/Psum ;
	*S4 += P4/Psum ;
	*S5 += P5/Psum ;
}

/*
 *	current is the weighted average of the five set of counters 
 */
static weightedaverage(current, D, H, V, A, B) 
	FScell *current, *H, *D, *V, *A, *B ;
{
	float PD, PH, PV, PA, PB, Psum ;

	scaledprobability5(-current->D,-current->H,-current->V,-current->A,-current->B,&PD,&PH,&PV,&PA,&PB,&Psum) ;

	current->DM = (D->DM*PD + H->DM*PH + V->DM*PV + A->DM*PA + B->DM*PB)/Psum ;
	current->DC = (D->DC*PD + H->DC*PH + V->DC*PV + A->DC*PA + B->DC*PB)/Psum ;
	current->DH = (D->DH*PD + H->DH*PH + V->DH*PV + A->DH*PA + B->DH*PB)/Psum ;
	current->DV = (D->DV*PD + H->DV*PH + V->DV*PV + A->DV*PA + B->DV*PB)/Psum ;
	current->DA = (D->DA*PD + H->DA*PH + V->DA*PV + A->DA*PA + B->DA*PB)/Psum ;
	current->DB = (D->DB*PD + H->DB*PH + V->DB*PV + A->DB*PA + B->DB*PB)/Psum ;

	current->HM = (D->HM*PD + H->HM*PH + V->HM*PV + A->HM*PA + B->HM*PB)/Psum ;
	current->HC = (D->HC*PD + H->HC*PH + V->HC*PV + A->HC*PA + B->HC*PB)/Psum ;
	current->HH = (D->HH*PD + H->HH*PH + V->HH*PV + A->HH*PA + B->HH*PB)/Psum ;
	current->HV = (D->HV*PD + H->HV*PH + V->HV*PV + A->HV*PA + B->HV*PB)/Psum ;

	current->VM = (D->VM*PD + H->VM*PH + V->VM*PV + A->VM*PA + B->VM*PB)/Psum ;
	current->VC = (D->VC*PD + H->VC*PH + V->VC*PV + A->VC*PA + B->VC*PB)/Psum ;
	current->VH = (D->VH*PD + H->VH*PH + V->VH*PV + A->VH*PA + B->VH*PB)/Psum ;
	current->VV = (D->VV*PD + H->VV*PH + V->VV*PV + A->VV*PA + B->VV*PB)/Psum ;

	current->AD = (D->AD*PD + H->AD*PH + V->AD*PV + A->AD*PA + B->AD*PB)/Psum ;
	current->AA = (D->AA*PD + H->AA*PH + V->AA*PV + A->AA*PA + B->AA*PB)/Psum ;

	current->BD = (D->BD*PD + H->BD*PH + V->BD*PV + A->BD*PA + B->BD*PB)/Psum ;
	current->BB = (D->BB*PD + H->BB*PH + V->BB*PV + A->BB*PA + B->BB*PB)/Psum ;
}

/*
 *	compute the current cell from the three adjacent incomming cell
 *	this routine depends on the global variables costDH, costHH, ... etc
 */
static FS_ComputeCurrentCell(current,hori,ddiag,vert,match)
	FScell *current, *hori, *ddiag, *vert ;
	bool match ; /* true if the current characters matches */
{
	FScell H, D, V, A, B ;
	float CDD, CDH, CDV, CDA, CDB ; /* costs from anywhere to everywhere else */
	float CHD, CHH, CHV ;
	float CVD, CVH, CVV ;
	float CAD, CAH, CAV, CAA, CAB ;
	float CBD, CBH, CBV, CBA, CBB ;

	D = *ddiag ;
	H = B = *hori ;
	V = A = *vert ;

	/* the costs of entrance from the diagonal direction */
	if (match) {
		CDD = D.D+costDM ;
		CHD = D.H+costHM  ;
		CVD = D.V+costVM  ;
		CAD = D.A+costAD + costDM ;
		CBD = D.B+costBD + costDM ;
	} else {
		CDD = D.D+costDC ;
		CHD = D.H+costHC  ;
		CVD = D.V+costVC  ;
		CAD = D.A+costAD + costDC ;
		CBD = D.B+costBD + costDC ;
	}
	/* short horizontal */
	CDH = H.D+costDH ;
	CHH = H.H+costHH ;
	CVH = H.V+costVH ;
	CAH = H.A+costAD+costDH ;
	CBH = H.B+costBD+costDH ;

	/* short vertical */
	CDV = V.D+costDV ;
	CHV = V.H+costHV ;
	CVV = V.V+costVV ;
	CAV = V.A+costAD+costDV ;
	CBV = V.B+costBD+costDV ;

	/* long horizontal */
	CDB = H.D+costDB ;
	CAB = H.A+costAD + costDB ;
	CBB = H.B+costBB ;

	/* long vertical */
	CDA = V.D+costDA ;
	CAA = V.A+costAA ;
	CBA = V.B+costBD + costDA ;

	if (match) {
		weightedincrement(-CDD,-CHD,-CVD,-CAD,-CBD,&D.DM,&D.HM,&D.VM,&D.AD,&D.BD) ;
		weightedincrement2(-CVV,-CDV,-CHV,-CAV,-CBV,&V.VV,&V.DV,&V.HM,&V.AD,&V.BD) ;
		weightedincrement2(-CHH,-CDH,-CVH,-CAH,-CBH,&H.HH,&H.DH,&H.VM,&H.AD,&H.BD) ;
	} else {
		weightedincrement(-CDD,-CHD,-CVD,-CAD,-CBD,&D.DC,&D.HC,&D.VC,&D.AD,&D.BD) ;
		weightedincrement2(-CVV,-CDV,-CHV,-CAV,-CBV,&V.VV,&V.DV,&V.HC,&V.AD,&V.BD) ;
		weightedincrement2(-CHH,-CDH,-CVH,-CAH,-CBH,&H.HH,&H.DH,&H.VC,&H.AD,&H.BD) ;
	}
	weightedincrement3(-CAA,-CDA,-CBA,&A.AA,&A.DA,&A.BD) ;
	weightedincrement3(-CBB,-CDB,-CAB,&B.BB,&B.DB,&B.AD) ;

	current->H = sumlg5(CDH, CHH, CVH, CAH, CBH) ;
	current->V = sumlg5(CDV, CHV, CVV, CAV, CBV) ;
	current->D = sumlg5(CDD, CHD, CVD, CAD, CBD) ;
	current->A = sumlg3(CDA, CAA, CBA) ;
	current->B = sumlg3(CDB, CAB, CBB) ;

	weightedaverage(current,&D,&H,&V,&A,&B) ;
}

/*
 *	compute the various cost from adaptive coding
 */
static FS_AdaptiveCost(left,diag,top,match)
	FScell *left,*diag,*top ;
	bool match ;
{
	float PDM,PDC,PDH,PDV,PDA,PDB ;
	float PHM,PHC,PHH,PHV ;
	float PVM,PVC,PVH,PVV ;
	float PAD,PAA,PBD,PBB ;
	float log12 ;

	log12 = (float) log2((double)12) ;

#ifdef FREEADAPT
	/*
	 *	Let all the probs to adapt freely ... tends to give too high a prob
	 *	to the long indels
	 */
	if (match) {
		PDM = (diag->DM+1)/(sumD(diag)+5) ;
		PHM = (diag->HM+1)/(sumH(diag)+4) ; 
		PVM = (diag->VM+1)/(sumV(diag)+4) ; 
		costDM  = (float)log2((double)PDM) - 2 ;
		costHM  = (float)log2((double)PHM) - 2 ;
		costVM  = (float)log2((double)PVM) - 2 ;
	} else {
		PDC = (diag->DC+1)/(sumD(diag)+5) ;
		PHC = (diag->HC+1)/(sumH(diag)+4) ; 
		PVC = (diag->VC+1)/(sumV(diag)+4) ; 
		costDC  = (float)log2((double)PDC) - log12 ;
		costHC  = (float)log2((double)PHC) - log12 ;
		costVC  = (float)log2((double)PVC) - log12 ;
	}
	PBD = (diag->BD+1)/(sumB(diag)+2) ; 
	PAD = (diag->AD+1)/(sumA(diag)+2) ; 

	PDH = (left->DH+1)/(sumD(left)+5) ; 
	PHH = (left->HH+1)/(sumH(left)+4) ; 
	PVH = (left->VH+1)/(sumV(left)+4) ; 

	PDV = (top->DV+1)/(sumD(top)+5) ; 
	PHV = (top->HV+1)/(sumH(top)+4) ; 
	PVV = (top->VV+1)/(sumV(top)+4) ; 

	PDB = (left->DB+1)/(sumD(left)+5) ; 
	PBB = (left->BB+1)/(sumB(left)+2) ; 

	PDA = (top->DA+1)/(sumD(top)+5) ; 
	PAA = (top->AA+1)/(sumA(top)+2) ; 

#else

	/*
	 *	fix the probs for the long indels
	 */
	PDM = (diag->DM+1)/(diag->DM+diag->DC+diag->DH+diag->DV+4) - 0.002;
	PHM = (diag->HM+1)/(sumH(diag)+4) ; 
	PVM = (diag->VM+1)/(sumV(diag)+4) ; 
	costDM  = (float)log2((double)PDM) - 2 ;
	costHM  = (float)log2((double)PHM) - 2 ;
	costVM  = (float)log2((double)PVM) - 2 ;

	PDC = (diag->DC+1)/(diag->DM+diag->DC+diag->DH+diag->DV+4) ;
	PHC = (diag->HC+1)/(sumH(diag)+4) ; 
	PVC = (diag->VC+1)/(sumV(diag)+4) ; 
	costDC  = (float)log2((double)PDC) - log12 ;
	costHC  = (float)log2((double)PHC) - log12 ;
	costVC  = (float)log2((double)PVC) - log12 ;

	PDA = PDB = 0.001 ;
	PAD = PBD = 0.01 ;
	PAA = PBB = 0.99 ;

	PDH = (left->DH+1)/(left->DM+left->DC+left->DH+left->DV+4) ;
	PHH = (left->HH+1)/(sumH(left)+4) ; 
	PVH = (left->VH+1)/(sumV(left)+4) ; 

	PDV = (top->DV+1)/(top->DM+top->DC+top->DH+top->DV+4) ;
	PHV = (top->HV+1)/(sumH(top)+4) ; 
	PVV = (top->VV+1)/(sumV(top)+4) ; 

#endif
	/*--------------------------------------------*/

	costBD  = (float)log2((double)PBD) - 2 ;
	costAD  = (float)log2((double)PAD) - 2 ;

	costDH  = (float)log2((double)PDH) - 2 ;
	costHH  = (float)log2((double)PHH) - 2 ;
	costVH  = (float)log2((double)PVH) - 2 ;

	costDV  = (float)log2((double)PDV) - 2 ;
	costHV  = (float)log2((double)PHV) - 2 ;
	costVV  = (float)log2((double)PVV) - 2 ;

	costDB  = (float)log2((double)PDB) - 2 ;
	costBB  = (float)log2((double)PBB) - 2 ;

	costDA  = (float)log2((double)PDA) - 2 ;
	costAA  = (float)log2((double)PAA) - 2 ;
}

/*
 *	The last argument is the forward matrix.
 *	The calling routine can choose to supply NULL to this argument
 *	when the matrix is not required.
 */
static DPA( mode, P, result, matrix)
	int  mode ;
	FScell *result ;
	FiveStatePBlock *P ;
	FSpcell matrix[][MATRIXSIZE] ;
{
	FScell c1[MAXSEQLENGTH], c2[MAXSEQLENGTH] ;
	FScell *d1, *d2, *temp, boundarycell ;
	int i, j ;

	FS_initcell(&boundarycell) ;
	if (mode == fixed) FS_ProbToCosts(P);

	d1 = c1 ;
	d2 = c2 ;
	/* Boundary Conditions */
	FS_initcell(&d1[0]) ;
	d1[0].D = (float)log2((double)P->D) ;
	d1[0].H = (float)log2((double)P->H) ;
	d1[0].V = (float)log2((double)P->V) ;
	d1[0].A = (float)log2((double)P->A) ;
	d1[0].B = (float)log2((double)P->B) ;
	if (matrix != NULL) copy(&matrix[0][0],&d1[0]) ;
	debug(TS_printcell(0,0,&d1[0]) ;)

	for (j = 1 ; j <= n; j++) {
		if (mode == adaptive)
			FS_AdaptiveCost(&d1[j-1],&boundarycell,&boundarycell,FALSE) ;
		FS_ComputeCurrentCell(&d1[j],&d1[j-1],&boundarycell,&boundarycell,FALSE) ;
		if (matrix != NULL) copy(&matrix[0][j],&d1[j]) ;
		debug(TS_printcell(0,j,&d1[j]) ;)
	}
 
	for (i = 1 ; i <= m ; i++) { /* General Step */
		if (mode == adaptive)
			FS_AdaptiveCost(&boundarycell,&boundarycell,&d1[0],FALSE) ;
		FS_ComputeCurrentCell(&d2[0],&boundarycell,&boundarycell,&d1[0],FALSE);
		if (matrix != NULL) copy(&matrix[i][0],&d2[0]) ;
		debug(TS_printcell(i,0,&d2[0]) ;)

		for (j = 1; j <= n; j++) {
			if (mode == adaptive )
				FS_AdaptiveCost(&d2[j-1],&d1[j-1],&d1[j],a[i]==b[j]) ;
			FS_ComputeCurrentCell(&d2[j],&d2[j-1],&d1[j-1],&d1[j],a[i]==b[j]) ;
			if (matrix != NULL) copy(&matrix[i][j],&d2[j]) ;
			debug(TS_printcell(i,j,&d2[j]) ;)
		} /* for j */
		swap(d1, d2) ;
	} /* for i */
	*result = d1[n] ;
} /*DPA*/

static FS_ComputePCell(current,horz,diag,vert,charequal)
	FSpcell *current,*horz,*diag,*vert ;
	bool charequal ;
{
	float costD ;

	costD = (charequal?costDM:costDC) ;
	current->D = sumlg5(
			costD  + diag->D,
			costDH + horz->H,
			costDB + horz->B,
		 	costDA + vert->A,
			costDV + vert->V ) ;
	current->H = sumlg3(
			(charequal?costHM:costHC) + diag->D,
			costHH + horz->H,
			costHV + vert->V) ;
	current->B = sumlg5(
			costBD + costD  + diag->D,
			costBD + costDH + horz->H,
			         costBB + horz->B,
			costBD + costDV + vert->V,
			costBD + costDA + vert->A ) ;
	current->V = sumlg3(
			(charequal?costVM:costVC) + diag->D,
			costVH + horz->H,
			costVV + vert->V) ;
	current->A = sumlg5(
			costAD + costD  + diag->D,
			costAD + costDH + horz->H,
			costAD + costDB + horz->B,
			costAD + costDV + vert->V,
			         costAA + vert->A ) ;
}
/*
 *	Fwd[i][j] given by DPA gives the prob that the machine ends in the
 *	various states after generating A[1..i], B[1..j].
 *	This routine computes the matrix Rev[i][j] which gives the prob
 *	that FROM the various starting state the machine will
 *	generate A[i+1..m], B[j+1..n].
 *	So Fwd[i][j] + Rev[i][j] i=1..m, j = 1..n gives the density plot.
 */
static DPArev(P,matrix)
	FiveStatePBlock *P ;
	FSpcell matrix[][MATRIXSIZE] ;
{
	FSpcell e1[MAXSEQLENGTH], e2[MAXSEQLENGTH] ;
	FSpcell *d1, *d2, *temp, *horz, *vert, *diag, boundarycell ;
	int i, j ;

	FS_initpcell(&boundarycell) ;
	FS_ProbToCosts(P) ;

	d1 = e1 ;
	d2 = e2 ;
	/* Boundary Conditions */
	d1[n+1].D = d1[n+1].H = d1[n+1].V = d1[n+1].A = d1[n+1].B = 0.0 ;
	matrix[m+1][n+1] = d1[n+1] ;
      
	for (j = n ; j >= 1; j--) {
		FS_ComputePCell(&d1[j],&d1[j+1],&boundarycell,&boundarycell,FALSE) ;
		matrix[m+1][j] = d1[j] ;
	}
 
	for (i = m ; i >= 1 ; i--) { /* General Step */
		FS_ComputePCell(&d2[n+1],&boundarycell,&boundarycell,&d1[n+1],FALSE) ;
		matrix[i][n+1] = d2[n+1] ;
		for (j = n; j >= 1; j--) {
			FS_ComputePCell(&d2[j],&d2[j+1],&d1[j+1],&d1[j],a[i]==b[j]) ;
			matrix[i][j] = d2[j] ;
		}
		swap(d1, d2) ;
	}
} /*DPArev*/

/*------------------------------------------------------------------------------
 *	Computes the 5 states R-theory message length with a fixed set of parameters
 */
FixParmFiveState(A,B,P,result)
	string *A, *B ;
	FiveStatePBlock *P ;
	FScell *result ;
{
	a = A->sequence ; b = B->sequence ;
	m = A->length ; n = B->length ;
	DPA(fixed,P,result,(FSpcell *) NULL) ;
}
