
/*
 *  Copyright  (c)  L.Allison, C.S.Wallace, C.N.Yee.
 *     Dept. Computer Science, Monash University, Australia 3168,    May 1991.
 *  1. These routines may only be used for non-commercial, non-classified,
 *     research purposes.
 *  2. No liability of any kind is accepted for their use.
 *  3. As a courtesy, please inform us of any interesting uses of the routines,
 *     of any bugs that you find in them and of how to correct
 *     such bugs, if known.
 *  4. Do not delete this copyright notice.
 *  5. Please include the following references in any publication based wholly
 *     or in part on the use of these routines or on the ideas contained in them:
 *   . Allison L. & C.N.Yee.
 *        Minimum Message Length Encoding and the Comparison of Macro-Molecules.
 *        Bull. Math. Bio. 52(3) 431-453 1990
 *   . Allison L., C.S.Wallace & C.N.Yee. Finite-State Models in the Alignment
 *        of Macromolecules. J. Mol. Evol. (1992) 35:77-89
 *  6. You may distribute the routines to others if and only if
 *     they agree to accept and maintain these conditions and no fee is charged.
 *  7. There is no rule 7.
 */
/*
 *	"IO.c" Strings IO and miscellaneious routines
 *
 *	Note :
 *		index for a sequence is 1..length.  element 0 contains NULLSymbol
 *		index for an alignment is 1..length
 *
 *	Entry Points :
 *
 *	int ReadSequence(stream, s, length, Table) 
 *	FILE	*stream ;
 *	symbol	*seq[] ;
 *	int	*length ;
 *	AlphabetTable *Table ;
 *	
 *		read in a sequence from STREAM.
 *		The routine will allocate the required space for the sequence
 *		whose address will be returned in S.
 *		The length of the sequence is returned in LENGTH
 *		Also do internal symbol conversion and error checking
 *
 *	ReadString(f,A,Table)
 *	FILE *f ;
 *	string *A ;
 *	AlphabetTable *Table ;
 *
 *		read in a string -- a line of identifier, a line of description
 *		and a line of sequence
 *
 *	GenbankReadString(f,A,Table)
 *	FILE *f ;
 *	string *A ;
 *	AlphabetTable *Table ;
 *
 *		read in a string in genbank(?) format -- a line that begin with the
 *		delimiter '>' followed by an identifier, a line of description,
 *		then the sequence, in possibly more than one line.
 *
 *	TranslateToSymbol(c, sequeuce, Table)
 *	char	c[] ;
 *	symbol	*sequence[] ;
 *	AlphabetTable	*Table ;
 *
 *		translate a character array into a sequence according to
 *		alphabet table TABLE
 *
 *
 *	fDisplaySequence(stream, s, length, Table)
 *	DisplaySequence(s, length, Table)
 *	fPrintSequence(stream, s, length, Table)
 *	PrintSequence(s, length, Table)
 *
 *	symbol	*s ;
 *	int	length ;
 *	AlphabetTable *Table ;
 *
 *		all these routine output the sequence s.
 *		DisplaySequence will break the sequence into LINELENGTH characters
 *		per line.
 *
 *	PrintString(s, Table)
 *	string *s ;
 *	AlphabetTable *Table ;
 *
 *		print a string (identifier, description and sequence)
 *
 *	PrintAlignment(Alignment, AlignLength, A, B, Table) 
 *	align	Alignment[] ;
 *	symbol	A[], B[] ;
 *	int	AlignLength ;
 *	AplhabetTable	*Table ;
 *
 *	Print3DAlignment(Alignment, AlignLength, A, B, C, Table) 
 *	align3D	Alignment[] ;
 *	symbol	A[], B[], C[] ;
 *	int	AlignLength ;
 *	AplhabetTable	*Table ;
 *
 *	RandomString(s, Table)
 *	string *s ;
 *	AlphabetTable *Table ;
 *		
 *		allocate space for and generates a random sequence of
 *		length s->length
 *
 *	RandomMutation(a, b, Mutations, PChanges, PInsertions, PDeletions, Table)
 *	string *a, *b ;
 *	int	Mutations, PChanges, PInsertions, PDeletions ;
 *	AlphabetTable	*Table ;
 *		
 *		do a Mutations number of random mutation on the sequence a.
 *		the resultant sequence is created and its pointer returned in b.
 *		PChanges, PInsertions and PDeletions are the ralative
 *		frequencies for Changes, Insertions and Deletions respectively 
 *
 *	StrConstantLengthMutation(a, b, Mutations, PChanges, PInsDel, Table)
 *		same a RandomMutation(),  execpt that the length of the mutated
 *		sequence is kept the same as the original sequence.
 *		PInsDel is now the relative frequency or getting a Insetion
 *		or a deletion.
 *
 *	SeqConstantLengthMutation(a, b, len, Mutations, PChanges, PInsDel, Table)
 *	symbol	a[], b[] ;
 *		same as StrConstantLengthMutation() except at the sequence level.
 *		Note : the space for b is NOT allocated by the routine.
 *
 *	ReverseSequence(s, length)
 *	symbol	s[] ;
 *	int	length ;
 *		reverse a sequence in situ
 *
 *	ReverseAlignment(a, length)
 *	symbol	a[] ;
 *	int	length ;
 *		reverse an alignment in situ
 */
#include "macro.h"
#include <ctype.h>
#include <sys/time.h>

#include "Strings.h"
#include "error.h"

AlphabetTable DNASymbol = { 4, {NULLSymbol, 'A', 'C', 'G', 'T'} } ;
ComplementTable DNAComplement = { 4, {0, 4, 3, 2, 1} } ; /* {T,G,C,A} */

AlphabetTable AlphaSymbol = { 53,
		{ NULLSymbol, ' ',
		  'A','B','C','D','E','F','G','H','I','J','K','L','M',
		  'N','O','P','Q','R','S','T','U','V','W','X','Y','Z',
		  'a','b','c','d','e','f','g','h','i','j','k','l','m',
		  'n','o','p','q','r','s','t','u','v','w','x','y','z'
		}
	} ;
 
AlphabetTable *SymbTable = &DNASymbol ;

char 	*malloc() ;
long 	time() ;

/*
 *	This cost function minimises the number of changes
 */
float Cost11(a, b) 
	symbol	a, b ;
{
	if (a == b) return 0.0 ;
	else	return 1.0 ;
}

/*
 *	maximise the number of matches --- LCS
 */
float Cost21(a, b)
	symbol	a, b ;
{
	if (a == b) return 0.0 ;
	else if ((a == NULLSymbol) || (b == NULLSymbol)) return 1.0 ;
	else	return 2.0 ;
}

printline(line,length)
	char	line[] ;
	int	length ;
{
	int i ;

	for (i = 0; i < length ; i++) putchar(line[i]) ;
	putchar('\n') ;
}

print3lines(line1, line2, line3, length)
	char	line1[], line2[], line3[] ;
	int	length ;
{
	printline(line1, length) ;
	printline(line2, length) ;
	printline(line3, length) ;
	putchar('\n') ;
}

readln(f)
	FILE	*f ;
{
	while (getc(f) != '\n') ;
}


/*
 *	read a line(delimeted by '\n' or EOF) from stream f into array s
 *	also check that current line is not too long for s
 *	return FALSE if EOF is reached
 */
bool getline(f, s, maxlen)
	char s[] ;
	FILE *f ;
	int maxlen ;
{
	int	len ;

	len = 0 ;
	while (((s[len] = (char) getc(f)) != '\n') && ((int) s[len] != EOF)) {
		if (++len >= maxlen) {
			(*ErrorHandling[ERRLineTooLong])(ERRLineTooLong) ;
			return FALSE ;
		}
	}

	if ( (int) s[len] == EOF ) return FALSE ;
	
	s[len] = NULL ;
	return TRUE ;
}

bool ToSymbol(c, symb, Table)
	char	c ;
	symbol *symb ;
 	AlphabetTable *Table ;
{
	int 	i ;

	/* can change to binary search later on */
	for ( i = 1 ; i <= Table->size ; i ++)
		if (Table->Symbol[i] == c) {*symb = i ; return TRUE ;}
	
	return FALSE ;
}

ReadSequence(f, seq, length, Table) 
	FILE	*f ;
	symbol	*seq[] ;
	int	*length ;
 	AlphabetTable *Table ;
{
	int 	i, len, s[MAXSEQLENGTH] ;
	int		badchars[256] ; /* frequencies of occurrances of illegal symbols */
	bool	foundbadchars ;
	char	c, c1 ; 

	len = 0 ;
	foundbadchars = FALSE ;
	for (i=0 ; i<=255; i++) badchars[i] = 0 ;

 	while ( (c = getc(f)) != '\n') {
		if (islower(c)) c1 = toupper(c) ; else c1 = c ;
		if (!ToSymbol(c1, &s[++len], Table)) {
			/* found a illegal character */
			len-- ; /* skip it */
			if (c != ' ') {
				badchars[c] ++ ;
				foundbadchars = TRUE ;
			}
		}
    	if (len >= MAXSEQLENGTH) {
			(*ErrorHandling[ERRLineTooLong])(ERRLineTooLong) ;
			return ;
		}
	}
	*seq = (symbol *) malloc((unsigned) (len+1) * sizeof(symbol)) ;
	(*seq)[0] = NULLSymbol ;

	for ( i = 1; i <= len ; i++) (*seq)[i] = s[i] ;

	*length = len ;

	if (foundbadchars) {
		printf("Warning : illegal characters ignored\n") ;

		for (i = 0; i <= 255 ; i++)
			if (badchars[i])
				printf("char : %c    frequency: %d\n",i,badchars[i]) ;
	}
} /* ReadSequence */


ReadString(f,A,Table)
	FILE *f ;
	string *A ;
	AlphabetTable *Table ;
{
	getline(f, A->identifier,IDLength) ;
	getline(f, A->description,DESCLength) ;
	ReadSequence(f,&A->sequence, &A->length, Table) ;
}

GenbankReadSequence(f, seq, length, Table) 
	FILE	*f ;
	symbol	*seq[] ;
	int	*length ;
 	AlphabetTable *Table ;
{
	int 	i, len, s[MAXSEQLENGTH] ;
	char	c, c1 ; 
	int		badchars[256] ; /* frequencies of occurrances of illegal symbols */
	bool	foundbadchars ;

	len = 0 ;
	foundbadchars = FALSE ;
	for (i=0 ; i<=255; i++) badchars[i] = 0 ;

 	while ( ((c = getc(f)) != GENBANKDELIMITER) && (c != EOF)) {
		if (c == '\n' || c == ' ') continue ;
		if (islower(c)) c1 = toupper(c) ; else c1 = c ;
		if (!ToSymbol(c1, &s[++len], Table)) { /* found a illegal character */
			len-- ; /* skip it */
			if (c != ' ') {
				badchars[c] ++ ;
				foundbadchars = TRUE ;
			}
		}

    	if (len >= MAXSEQLENGTH) {
			(*ErrorHandling[ERRLineTooLong])(ERRLineTooLong) ;
			return ;
		}
	}

	*seq = (symbol *) malloc((unsigned) (len+1) * sizeof(symbol)) ;
	(*seq)[0] = NULLSymbol ;

	for ( i = 1; i <= len ; i++) (*seq)[i] = s[i] ;

	*length = len ;

	if (foundbadchars) {
		printf("Warning : illegal characters ignored\n") ;

		for (i = 0; i <= 255 ; i++)
			if (badchars[i])
				printf("char : %c    frequency: %d\n",i,badchars[i]) ;
	}
} /* GenbankReadSequence */


GenbankReadString(f,A,Table)
	FILE *f ;
	string *A ;
	AlphabetTable *Table ;
{
	int c ;

	if ((c = getc(f)) != GENBANKDELIMITER) ungetc(c,f) ;
	getline(f, A->identifier,IDLength) ;
	getline(f, A->description,DESCLength) ;
	GenbankReadSequence(f,&A->sequence, &A->length, Table) ;
}


/*
 *	translate a character array into a sequence according to
 *	alphabet table TABLE
 */
TranslateToSymbol(c, seq, Table)
	char	c[] ;
	symbol	*seq[] ;
	AlphabetTable	*Table ;
{
	int	i, len ;
	symbol	s[MAXSEQLENGTH] ;

	len = 0 ;

 	while ( c[++len] ) ToSymbol(c[len], &s[len], Table) ;

	*seq = (symbol *) malloc((unsigned) (len) * sizeof(symbol)) ;

	(*seq)[0] = 0 ;
	for ( i = 1; i <= len ; i++)
		(*seq)[i] = s[i] ;
}

/***************************************************************************/

fPrintSequence(stream, s, length, Table)
	FILE *stream ;
	symbol	s[] ;
	int	length ;
 	AlphabetTable *Table ;
{
	int	i ;

	for (i = 1 ; i <= length; i++) putc(Table->Symbol[s[i]],stream) ;
	putc('\n',stream) ;
}

PrintSequence(s, length, Table)
	symbol	s[] ;
	int	length ;
 	AlphabetTable *Table ;
{	fPrintSequence(stdout, s, length, Table) ; }



DisplaySequence(s, length, Table)
	symbol	s[] ;
	int	length ;
 	AlphabetTable *Table ;
{	fDisplaySequence(stdout, s, length, Table) ; }

fDisplaySequence(stream, s, length, Table)
	FILE *stream ;
	symbol	s[] ;
	int	length ;
 	AlphabetTable *Table ;
{
	int	i, count;

	count = 0 ;
	for (i = 1 ; i <= length; i++) {
		if (++count > LINELENGTH) {
			count = 0 ;
			putc('\n',stream) ;
		}
		putc(Table->Symbol[s[i]],stream) ;
	}
	putc('\n',stream) ;
}

PrintString(s, Table)
	string *s ;  
 	AlphabetTable *Table ;
{
	printf("%s        %d bases\n", s->identifier,s->length) ;
	printf("%s\n", s->description) ;
	DisplaySequence(s->sequence, s->length, Table) ;
}

PrintAlignment(Alignment, AlignLength, A, B, Table) 
	align	Alignment[] ;
	symbol	A[], B[] ;
	int	AlignLength ;
	AlphabetTable	*Table ;
{
	int	i, lo, hi ;

	hi = 0 ;
	while (hi < AlignLength) {
		lo = hi+1 ;
		hi = min(AlignLength, hi + LINELENGTH) ;
		for ( i = lo ; i <= hi; i++)
			if (Alignment[i].a == 0) putchar('-');
			else putchar(Table->Symbol[A[Alignment[i].a]]) ;
		putchar('\n') ;
		for ( i = lo ; i <= hi; i++)
			if ((Alignment[i].a != NULLSymbol) &&
				(Alignment[i].b != NULLSymbol) &&
				(A[Alignment[i].a] == B[Alignment[i].b])) putchar('|');
			else putchar(' ') ;

		putchar('\n') ;
		for ( i = lo ; i <= hi; i++)
			if (Alignment[i].b == 0) putchar('-');
			else putchar(Table->Symbol[B[Alignment[i].b]]) ;
		putchar('\n') ;
		putchar('\n') ;
	}
}

Print3DAlignment(Alignment, AlignLength, A, B, C, Table) 
	align3D	Alignment[] ;
	symbol	A[], B[], C[] ;
	int	AlignLength ;
	AlphabetTable	*Table ;
{
	char	line1[LINELENGTH], line2[LINELENGTH], line3[LINELENGTH];
	int	i, count ;

	count = 0 ;
	for (i = 1 ; i <= AlignLength ; i++) {
		if (Alignment[i].a == 0) line1[count] = '-' ;
		else line1[count] = Table->Symbol[A[Alignment[i].a]] ;

		if (Alignment[i].b == 0) line2[count] = '-' ;
		else line2[count] = Table->Symbol[B[Alignment[i].b]] ;

		if (Alignment[i].c == 0) line3[count] = '-' ;
		else line3[count] = Table->Symbol[C[Alignment[i].c]] ;

		if (++count > LINELENGTH) {
			count = 0 ;
			print3lines(line1,line2,line3,LINELENGTH) ;
		}
	}
	if (count != 0)
		print3lines(line1,line2,line3,count) ;
}

DuplicateSequence(a, alen, b, blen)
	symbol a[], *b[] ;
	int	alen, *blen ;
{
	int	i ;

	*b = (symbol *) malloc((unsigned) (alen+1) * sizeof(symbol)) ;

	for ( i = 0 ; i <= alen; i++) (*b)[i] = a[i] ;

	*blen = alen ;
}

symbol randomsymbol(Table)
	AlphabetTable *Table ;
{
	return 1 + randomnum(Table->size) ;
}

/*
 * generate a pair of random symbol that are different
 */
randomsymbolpair(Table,a,b)
	AlphabetTable *Table ;
	symbol *a, *b ;
{
	*a = 1+randomnum(Table->size) ;
	*b = 1+randomnum(Table->size-1) ;
	if (*b >= *a)
		(*b) ++ ;
}

/*
 *	generates a random sequence
 */
RandomSequence(s, length, Table) 
	symbol **s ;
	int	length ;
 	AlphabetTable *Table ;
{
	int	i ;

	*s = (symbol *) malloc((unsigned) (length+1) * sizeof(symbol)) ;
	(*s)[0] = NULLSymbol ;
	for (i = 1; i <= length; i++)
		(*s)[i]  = randomsymbol(Table) ;

}

RandomString(s, Table)
	string  *s ;
 	AlphabetTable *Table ;
{
	RandomSequence(&s->sequence, s->length, Table) ;
}

/*
 * do a random mutation
 */
RandomMutation(a, b, Mutations, PChanges, PInsertion, PDeletion, Table)
	string *a, *b ;
	int	Mutations, PChanges, PInsertion, PDeletion ;
	AlphabetTable	*Table ;
{
	symbol t[MAXSEQLENGTH] ;
	int	i, j, choice, pos ;
	int	P, PI, len ;
	
	P = PChanges + PInsertion + PDeletion ;
	PI = PChanges + PInsertion ;
	len = a->length ;

	for ( i = 1 ; i <= len; i++)
		t[i] = a->sequence[i] ;
	
	for ( i = 1 ; i <= Mutations ; i++) {
		choice = randomnum(P) ;
		if (choice < PChanges) { /* changes */
			if (len > 0) {
				pos = randomnum(len) + 1 ; /* 1..length */
				t[pos] = randomsymbol(Table) ;
				debug(printf("change at %d %d\n",pos, t[pos]) ;)
			}
		} else if ( choice < PI) {/* insertion */
			pos = 1 + randomnum(len+1)  ; /* 1..length+1 */
			for ( j = ++len; j > pos; j-- )
				t[j] = t[j-1] ;
			t[pos] = randomsymbol(Table) ;
			debug(printf("insertion at %d %d\n",pos, t[pos]) ;)
		} else { /* deletion */
			if (len > 0) {
				pos = randomnum(len) + 1 ; /* 1..length */
				debug(printf("deletion at %d %d\n",pos, t[pos]) ;)
				for ( j = pos; j < len ; j++)
					t[j] = t[j+1] ;
				len-- ;
			}
		}
	}

	DuplicateSequence(t, len, &b->sequence, &b->length) ;
}

StrConstantLengthMutation(a, b, Mutations, PChanges, PInsDel, Table)
	string *a, *b ;
	int	Mutations, PChanges, PInsDel ;
	AlphabetTable	*Table ;
{

	b->sequence = (symbol *)malloc((unsigned) (a->length+1) * sizeof(symbol) ) ;
	SeqConstantLengthMutation(&a->sequence[1], &b->sequence[1], a->length, Mutations, PChanges, PInsDel, Table) ;
}

SeqConstantLengthMutation(a, b, len, Mutations, PChanges, PInsDel, Table)
	symbol *a, *b ;
	int	Mutations, PChanges, PInsDel ;
	AlphabetTable	*Table ;
{
	int	i, j, choice, pos ;
	int	P, PI ;
	int	nchanges,  ninsdel ;
	symbol	t[MAXSEQLENGTH] ;
	
	P = PChanges + PInsDel ;

	for ( i = 0 ; i < len; i++)
		t[i] = a[i] ;
	
	nchanges = ninsdel = 0 ;

	for ( i = 1 ; i <= Mutations ; i++) {
		choice = randomnum(P) ;
		if (choice < PChanges) /* changes */
			nchanges++ ;
		else
			ninsdel++ ;
	}

	if (odd(ninsdel)) /* make it even */
		if (randomnum(2) && nchanges > 0) { /* toss a coin */
			/* inscrement it */
			ninsdel++ ; nchanges-- ;
		} else { /* decrease it */
			ninsdel-- ; nchanges++ ;
		}
	debug(printf("changes: %d ins/del %d\n",nchanges, ninsdel) ;) 
	P = nchanges + ninsdel ;
	PI = nchanges + (int)(ninsdel/2) ;

	for ( i = 1 ; i <= Mutations ; i++) {
		choice = randomnum(P) ;
		if (choice < nchanges) { /* changes */
			if (len > 0) {
				pos = randomnum(len) ;
				t[pos] = randomsymbol(Table) ;
				debug(printf("change at %d %d\n",pos, t[pos]) ;)
				nchanges -- ; P-- ; PI-- ;
			}
		} else if ( choice < PI) {/* insertion */
			pos = randomnum(len)  ;
			for ( j = ++len; j > pos; j-- )
				t[j] = t[j-1] ;
			t[pos] = randomsymbol(Table) ;
			debug(printf("insertion at %d %d\n",pos, t[pos]) ;)
			PI-- ; P-- ;
		} else { /* deletion */
			if (len > 0) {
				pos = randomnum(len) + 1 ;
				debug(printf("deletion at %d %d\n",pos, t[pos]) ;)
				for ( j = pos; j < len ; j++)
					t[j] = t[j+1] ;
				len-- ; P-- ;
			}
		}
	}
	for ( i = 0 ; i < len; i++) b[i] = t[i] ;
}

ReverseSequence(s, length)
	symbol	s[] ;
	int	length ;
{
	int	i, mid ;
	symbol	temp ;

	mid = length >> 1 ;
	for ( i = 1 ; i <= mid ; i++)
		swap(s[i], s[length-i+1]) ;
}

ReverseAlignment(alignment, length)
	align	alignment[] ;
	int	length ;
{
	int	i, mid ;
	align	temp ;

	mid = length >> 1 ;
	for ( i = 1 ; i <= mid ; i++)
		swap(alignment[i], alignment[length-i+1]) ;
}

CheckAlignment(alignment1,length1,alignment2,length2)
	align alignment1[], alignment2[] ;
	int	length1, length2 ;
{
	int	i ;

	if (length1 != length2) return FALSE ;
	for (i = 1 ; i<=length1; i++)
		if ((alignment1[i].a != alignment2[i].a)
			 || (alignment1[i].b != alignment2[i].b ))
		return FALSE ;
	return TRUE ;
}

static float findmin(D,m,n)
	float D[][MATRIXSIZE];
	int m,n ;
{
	int i, j ;
	float minval ;

	minval = MAXINT ;
	for (i=0; i <= m ; i++)
		for (j=0; j <= n ; j++)
			if (D[i][j] < minval) minval = D[i][j] ;
	return minval ;
}

/*
 *	scale the matrix down by the minimum entry
 */
scale(D,m,n)
	float D[][MATRIXSIZE];
	int m, n ;
{
	float minval ;
	int i, j ;

	minval = findmin(D,m,n) ;
	for (i=0; i <= m ; i++) 
		for (j=0; j <= n ; j++)
			D[i][j] -= minval ;
}

/*
 *	Plot density matrix on screen
 */
PlotMatrix(a,b,D,m,n,Table)
	float D[][MATRIXSIZE] ;
	symbol a[], b[] ;
	int m, n ;
	AlphabetTable *Table; 
{
	int i, j ;
	float ml ;
	char ch ;

	printf("\n   ") ;
	for (j=1 ; j<= n ; j++) {
		/*putchar(' ');*/
		putchar(Table->Symbol[b[j]]);
	}
	putchar('\n') ;;
	for (i=0 ; i<= m ; i++) {
		if (i > 0 ) printf("%c:",Table->Symbol[a[i]]) ;
		else printf("  ");
		for (j=0 ; j <=  n; j++) {
			ml = D[i][j] ;
			if (ml > 16) ch=' ' ;
			else if (ml>8) ch='.' ;
			else if (ml>4) ch='-' ;
			else if (ml>2) ch='+' ;
			else if (ml>1) ch='*' ;
			else ch='#' ;
			putchar(ch) ;
			/*putchar(' ') ;*/
		}
		putchar('\n') ;
	}
	putchar('\n') ;
	printf("KEY:  #   *   +   -   . \n") ;
	printf("   :    1   2   4   8   16  bits over mml\n") ;   
	putchar('\n') ;
}

OutputMatrix(A,B,D,ofile,Caption,Table) 
	string *A, *B ;
	float D[][MATRIXSIZE] ;
	char *Caption ;
	FILE *ofile ;
	AlphabetTable *Table ;
{
	int i, j ;

	fprintf(ofile,"%d %d\n",A->length, B->length) ;
	fDisplaySequence(ofile, A->sequence, A->length, Table) ;
	fDisplaySequence(ofile, B->sequence, B->length, Table) ;
	for ( i = 0 ; i <= A->length ; i++) {
		for ( j = 0 ; j <= B->length ; j++)
			fprintf(ofile,"%5.1f ", D[i][j]) ;
		putc('\n',ofile) ;
	}
	fprintf(ofile,"%s\n",Caption) ;
	fprintf(ofile,"%s vs %s\n",A->identifier,B->identifier) ;
} /*plot*/

