
/*
 *  Copyright  (c)  L.Allison, C.S.Wallace, C.N.Yee.
 *     Dept. Computer Science, Monash University, Australia 3168,    May 1991.
 *  1. These routines may only be used for non-commercial, non-classified,
 *     research purposes.
 *  2. No liability of any kind is accepted for their use.
 *  3. As a courtesy, please inform us of any interesting uses of the routines,
 *     of any bugs that you find in them and of how to correct
 *     such bugs, if known.
 *  4. Do not delete this copyright notice.
 *  5. Please include the following references in any publication based wholly
 *     or in part on the use of these routines or on the ideas contained in them:
 *   . Allison L. & C.N.Yee.
 *        Minimum Message Length Encoding and the Comparison of Macro-Molecules.
 *        Bull. Math. Bio. 52(3) 431-453 1990
 *   . Allison L., C.S.Wallace & C.N.Yee. Finite-State Models in the Alignment
 *        of Macromolecules. J. Mol. Evol. (1992) 35:77-89
 *  6. You may distribute the routines to others if and only if
 *     they agree to accept and maintain these conditions and no fee is charged.
 *  7. There is no rule 7.
 */
/*
 * "Driver.c"
 *
 *	driver for public release of R-theory alignment program
 */
#include <stdio.h>
#include "macro.h"
#include "Strings.h"
#include "rtheory.h"

#define ONESTATE		1
#define THREESTATE		2
#define FIVESTATE		3
#define ONESTATER0		4
#define THREESTATER0	5
#define ALL				6

#define Accurarcy 0.000005

FILE *fopen() ;
FILE *OSoutfile, *TSoutfile, *FSoutfile, *infile ;

char filename[IDLength] ;

string A, B ;
OneStatePBlock OSParameters ;
OneStatePBlock OSR0Parameters ;
ThreeStatePBlock TSParameters ;
ThreeStatePBlock TSR0Parameters ;
FiveStatePBlock FSParameters ;

char *
MachineName(Machine)
	int Machine ;
{
	switch (Machine) {
		case ONESTATE    : return "One-State" ;
		case THREESTATE  : return "Three-State" ;
		case FIVESTATE   : return "Five-State" ;
		case ONESTATER0  : return "One-State best alignment" ;
		case THREESTATER0: return "Three-State best alignment" ;
		default          : printf("ERROR : unknown machine") ; exit() ;
	}
}

PromptForStrings()
{
	int c ;

	printf("Input the file name where the strings are in\n") ;
	printf("or type <RET> to input strings from terminal : ") ;

	if ((c = getchar()) == '\n') {
		printf("input string 1:\n") ;
		ReadSequence(stdin, &A.sequence, &A.length, &DNASymbol) ;
		printf("input string 2:\n") ;
		ReadSequence(stdin, &B.sequence, &B.length, &DNASymbol) ;
	} else {
		ungetc(c,stdin) ;
		getline(stdin,filename,IDLength) ;
		if (!(infile = fopen(filename,"r"))) {
			printf("Error opening %s\n",filename) ;
			exit(99) ;
		}
		printf("\n")  ;
		GenbankReadString(infile, &A, &DNASymbol) ;
		PrintString(&A, &DNASymbol) ;
		printf("\n")  ;
		GenbankReadString(infile, &B, &DNASymbol) ;
		PrintString(&B, &DNASymbol) ;
	}
}


PromptOneFile(Machine,outfile)
	int Machine ;
	FILE **outfile ;
{
	int c ;

	while (TRUE) {
		printf("File name to output density plot matrix for %s machine?\n",MachineName(Machine)) ;
		printf("or type <RET> if density plot is not required : ") ;
		if ((c = getchar()) != '\n') {
			ungetc(c,stdin) ;
			getline(stdin,filename,IDLength) ;
			if (*outfile=fopen(filename,"w")) return ;
			printf("Error opening file: %s\n",filename) ;
		} else {
			*outfile = (FILE *) NULL ;
			return ;
		}
	}
}

PromptOutputFile(Machine)
	int Machine ;
{
	if (A.length > MATRIXSIZE || B.length > MATRIXSIZE) 
		printf("Sequences are too long ... no density plot generated\n") ;
	else {
		putchar('\n') ;
		switch(Machine) {
			case ONESTATE :
				PromptOneFile(ONESTATE,&OSoutfile) ;
				break ;
			case THREESTATE :
				PromptOneFile(THREESTATE,&TSoutfile) ;
				break ;
			case FIVESTATE :
				PromptOneFile(FIVESTATE,&FSoutfile) ;
				break ;
			case ALL :
				PromptOneFile(ONESTATE,&OSoutfile) ;
				PromptOneFile(THREESTATE,&TSoutfile) ;
				PromptOneFile(FIVESTATE,&FSoutfile) ;
				break ;
		}
	}
}

int PromptMachinesToRun()
{
	int c ;

	while (TRUE) {
		printf("\nDo you want to run:\n") ;
		printf("    1. One-State machine\n") ;
		printf("    2. Three-State machine\n") ;
		printf("    3. Five-State machine\n") ;
		printf("    4. One-State machine one (best) alignment\n") ;
		printf("    5. Three-State machine one (best) alignment\n") ;
		printf("    6. All machines\n") ;
		printf("  Input option : ") ;
		c = getchar() ; readln(stdin) ;

		switch (c) {
			case '1' : return ONESTATE ;
			case '2' : return THREESTATE ;
			case '3' : return FIVESTATE ;
			case '4' : return ONESTATER0 ;
			case '5' : return THREESTATER0 ;
			case '6' : return ALL;
		}
	}
}

InitOSParameters(P)
	OneStatePBlock *P ;
{
	P->match = 0.8 ;
	P->change = 0.1 ;
	P->indel = 0.1 ;
}

InitTSParameters(P)
	ThreeStatePBlock *P ;
{
	P->DM = 0.75 ;
	P->DC = 0.15 ;
	P->DV = P->DH = 0.05 ;

	P->HM = P->VM = 0.37 ;
	P->HC = P->VC = 0.10 ;
	P->HH = P->VV = 0.50 ;
	P->HV = P->VH = 0.03 ;
	P->D = P->H = P->V = 1.0/3 ;
}

InitFSParameters(P)
	FiveStatePBlock *P ;
{
	P->DM = 0.748 ;
	P->DC = 0.15 ;
	P->DV = P->DH = 0.05 ;
	P->DA = P->DB = 0.001 ;

	P->HM = P->VM = 0.37 ;
	P->HC = P->VC = 0.10 ;
	P->HH = P->VV = 0.50 ;
	P->HV = P->VH = 0.03 ;

	P->AA = P->BB = 0.99 ;
	P->AD = P->BD = 0.01 ;

	P->D = P->H = P->V = P->A = P->B = 0.2 ;
}

InitParameters(Machine)
	int Machine ;
{
	switch (Machine) {
		case ONESTATE :
			InitOSParameters(&OSParameters) ;
			break ;
		case ONESTATER0 :
			InitOSParameters(&OSR0Parameters) ;
			break ;
		case THREESTATE :
			InitTSParameters(&TSParameters) ;
			break ;
		case FIVESTATE :
			InitFSParameters(&FSParameters) ;
			break ;
		case THREESTATER0 :
			InitTSParameters(&TSR0Parameters) ;
			break ;
		case ALL :
			InitOSParameters(&OSParameters) ;
			InitTSParameters(&TSParameters) ;
			InitFSParameters(&FSParameters) ;
			InitTSParameters(&OSR0Parameters) ;
			InitTSParameters(&TSR0Parameters) ;
			break ;
	}
}

ReadOSParameters(P)
	OneStatePBlock *P ;
{
	printf("One-State Machine Parameters:\n") ;
	printf("  Input P(Match), P(Change) : ") ;
	scanf("%f %f",&P->match,&P->change) ;
	P->indel = 1 - P->match - P->change ;
	printf("Parameters read in:") ;
	OS_printparameters(P) ; putchar('\n') ;
	readln(stdin) ;
}

ReadTSParameters(P)
	ThreeStatePBlock *P ;
{
	printf("Three-State Machine:\n") ;
	printf("State D:\n") ;
	printf("   input PDM, PDC :") ;
	scanf("%f %f",&P->DM, &P->DC) ; readln(stdin) ;
	P->DH = P->DV = (1 - P->DM - P->DC)/2 ;

	printf("State H and V (indels):\n") ;
	printf("   input PHM, PHC, PHH :",&P->HM,&P->HC,&P->HH) ;
	scanf("%f %f %f",&P->HM,&P->HC,&P->HH) ; readln(stdin) ;
	P->VM = P->HM ; P->VC = P->HC ; P->VV = P->HH ;
	P->VH = P->HV = 1 - P->HM - P->HC - P->HH  ;

	P->D = P->H = P->V = 1/3.0 ;
	printf("Parameters read in:\n") ;
	TS_printparameters(P) ; putchar('\n') ;
}

ReadFSParameters(P)
	FiveStatePBlock *P ;
{
	printf("Five-State Machine:\n") ;
	printf("State D-\n") ;
	printf("   input PDM, PDC, PDH :") ;
	scanf("%f %f %f",&P->DM, &P->DC,&P->DH) ; readln(stdin) ;
	P->DV = P->DH ;
	P->DB = P->DA = (1 - P->DM - P->DC - P->DH - P->DV)/2 ;

	printf("State H and V(short indels):\n") ;
	printf("   input PHM, PHC, PHH :",&P->HM,&P->HC,&P->HH) ;
	scanf("%f %f %f",&P->HM,&P->HC,&P->HH) ; readln(stdin) ;
	P->HV = P->VH =  1 - P->HM - P->HC - P->HH ;
	P->VM = P->HM ; P->VC = P->HC ; P->VV = P->HH ;
	printf("State A and B :\n") ;
	printf("   input PAA :") ;
	scanf("%f",&P->AA) ; readln(stdin) ;
	P->BD = P->AD = 1 - P->AA ;
	P->BB = P->AA ;

	P->D = P->H = P->V = P->A = P->B = 0.2 ;
	printf("Parameters read in:\n") ; putchar('\n') ;
	FS_printparameters(P) ;
}

ReadParameters(Machine)
	int Machine ;
{
	switch (Machine) {
		case ONESTATE :
		case ONESTATER0 :
			ReadOSParameters(&OSParameters) ;
			break ;
		case THREESTATE :
		case THREESTATER0 :
			ReadTSParameters(&TSParameters) ;
			break ;
		case FIVESTATE :
			ReadFSParameters(&FSParameters) ;
			break ;
		case ALL :
			ReadOSParameters(&OSParameters) ;
			ReadTSParameters(&TSParameters) ;
			ReadFSParameters(&FSParameters) ;
			break ;
	}
}

PromptParameters(Machine,mode)
	int Machine, *mode ;
{
	int c ;

	do {
		printf("\nInitial parameters -- do you want to :\n") ;
		printf("    1. Do adaptive coding on 1st run,  or\n") ;
		printf("    2. Use default parameters,  or\n") ;
		printf("    3. Input your own\n") ;
		printf("  Input option : ") ;
		c = getc(stdin); readln(stdin) ;
		if ((Machine == ONESTATER0 || Machine == THREESTATER0) && c == '1') {
			printf("Options 1 not valid...\n") ;
			c = '4' ;
		}
	} while ((c < '1') || (c > '3')) ;

	switch (c) {
		case '1' :	*mode = adaptive ;
					if (Machine == ALL) {
						InitParameters(ONESTATER0) ;
						InitParameters(THREESTATER0) ;
					}
					break ;
		case '2' :	*mode = fixed ;
					InitParameters(Machine) ;
					break ;
		case '3' :	*mode = fixed ;
					ReadParameters(Machine) ;
					break ;
	}
}


RunMachines(Machine,mode)
	int Machine, mode ;
{
	float	Null_ML,TS_ML,FS_ML,OS_ML, OSR0_ML, TSR0_ML ;
	align	Alignment[MAXALIGNLENGTH] ;
	int		Alignlength ;
	float	OSmatrix, TSmatrix, FSmatrix ;
	int		sumlength ;
	int 	Niteration ;

	sumlength = A.length+B.length ;

	if (Machine == ONESTATE || Machine == ALL) {
		printf("\nONE-STATE MACHINE :\n") ;
		OneStateML(&A,&B,&OS_ML,&OSmatrix,mode,OSoutfile,&OSParameters,Accurarcy);
		printf("1-State Machine: %1.4f(bits/symb) %6.2f(ML) %6.2f(Matrix)\n\n", OS_ML/sumlength,OS_ML,OSmatrix) ;
		fflush(stdout) ;
	}

	if (Machine == THREESTATE || Machine == ALL) {
		printf("\nTHREE-STATE MACHINE :\n");
		ThreeStateMLS(&A,&B,&TS_ML,&TSmatrix,mode,TSoutfile,&TSParameters,Accurarcy);
		printf("3-State Machine: %1.4f(bits/symb) %6.2f(ML) %6.2f(Matrix)\n\n", TS_ML/sumlength,TS_ML,TSmatrix) ;
		fflush(stdout) ;
	}

	if (Machine == FIVESTATE || Machine == ALL) {
		printf("\nFIVE-STATE MACHINE :\n") ;
		FiveState2ML(&A,&B,&FS_ML,&FSmatrix,mode,FSoutfile,&FSParameters,Accurarcy);
		printf("5-State Machine: %1.4f(bits/symb) %6.2f(ML) %6.2f(Matrix)\n\n", FS_ML/sumlength,FS_ML,FSmatrix) ;
		fflush(stdout) ;
	}

	if (Machine == ONESTATER0 || Machine == ALL) {
		printf("\nONE-STATE MACHINE :\n") ;
		ROptAlignString(&A,&B,Alignment,&Alignlength,&OSR0_ML,&Niteration,&OSR0Parameters,&DNASymbol) ;

 		PrintAlignment(Alignment, Alignlength, A.sequence, B.sequence, &DNASymbol) ;
		printf("1-State Machine best alignment: %1.4f(bits/symb) %6.2f(ML) \n\n", OSR0_ML/sumlength,OSR0_ML) ;
		fflush(stdout) ;
	}

	if (Machine == THREESTATER0 || Machine == ALL) {
		printf("\nTHREE-STATE MACHINE best alignment:\n");
		ThreeStateR0ML(&A,&B,Alignment,&Alignlength,&TSR0_ML,TSR0Parameters) ;

 		PrintAlignment(Alignment, Alignlength, A.sequence, B.sequence, &DNASymbol) ;
		printf("3-State Machine best alignment: %1.4f(bits/symb) %6.2f(ML)\n\n", TSR0_ML/sumlength,TSR0_ML,TSmatrix) ;
		fflush(stdout) ;
	}
	
	Null_ML = NullMesgLength(A.length,B.length) ;

	if (Machine == ALL) {
		printf("\nSUMMARY\n") ;
		printf("-------\n") ;
		printf("message length (bits/symbol) :\n") ;
		printf("Null: %1.3f OS: %1.3f(%1.3f) TS: %1.3f(%1.3f) FS: %1.3f(%1.3f) OS-R0: %1.3f TS-R0: %1.3f\n",
			Null_ML/sumlength, OS_ML/sumlength, OSmatrix/sumlength,
			TS_ML/sumlength,TSmatrix/sumlength,
			FS_ML/sumlength,FSmatrix/sumlength,
			OSR0_ML/sumlength,TSR0_ML/sumlength) ;
		putchar('\n') ;

		printf("message length (bits) :\n") ;
		printf("Null: %4.0f OS: %4.0f(%4.0f) TS: %4.0f(%4.0f) FS: %4.0f(%4.0f) OS-R0: %4.0f TS-R0: %4.0f\n",
			Null_ML, OS_ML, OSmatrix, TS_ML,TSmatrix, FS_ML,FSmatrix,OSR0_ML,TSR0_ML) ;
	} else
		printf("Null Theory : %1.3f(bits/symb) %4.0f(ML)\n",Null_ML/sumlength,Null_ML) ;
}

main()
{
	int Machine,mode ;

	PromptForStrings() ;
	Machine = PromptMachinesToRun() ;
	PromptOutputFile(Machine) ;
	PromptParameters(Machine,&mode) ;
	RunMachines(Machine,mode) ;
}
