
/*
 *  Copyright  (c)  L.Allison, C.S.Wallace, C.N.Yee.
 *     Dept. Computer Science, Monash University, Australia 3168,    May 1991.
 *  1. These routines may only be used for non-commercial, non-classified,
 *     research purposes.
 *  2. No liability of any kind is accepted for their use.
 *  3. As a courtesy, please inform us of any interesting uses of the routines,
 *     of any bugs that you find in them and of how to correct
 *     such bugs, if known.
 *  4. Do not delete this copyright notice.
 *  5. Please include the following references in any publication based wholly
 *     or in part on the use of these routines or on the ideas contained in them:
 *   . Allison L. & C.N.Yee.
 *        Minimum Message Length Encoding and the Comparison of Macro-Molecules.
 *        Bull. Math. Bio. 52(3) 431-453 1990
 *   . Allison L., C.S.Wallace & C.N.Yee. Finite-State Models in the Alignment
 *        of Macromolecules. J. Mol. Evol. (1992) 35:77-89
 *  6. You may distribute the routines to others if and only if
 *     they agree to accept and maintain these conditions and no fee is charged.
 *  7. There is no rule 7.
 */
/*
 *	"scale.c"
 *
 *	This program convert the matrix generated by the rtheory machines
 *	(c.f. 1State.c 3State.c, 5State.c) into a matrix of grey sale
 *	values 0..255.  This output from this program can be piped to
 *	pimg (c.f. printimage.c) which transform it into postscript 
 *	that can be display on a lazer writer
 *
 *	"usage: scale l(og) [contrast] | p | m [treshold]
 *
 *	The 'l' option is to display the matrix in logaritmetic in ML.
 *	The optional 'constrast' factor controls the sharpness of the
 *	high probability region.  The default value is 1. A high value
 *	(e.g. 4) gives a sharper picture while a low value (e.g. 0.2)
 *	includes more ML's in the high intensity region.
 *
 *	The 'p' option displays the image linear in probability (exp2 of ML).
 *
 *	The 'm' option display the image linear in ML. If the optional 'treshold'
 *	value is supplied then the ground (or 'white')level is set to
 *	min ML + treshold.  All values higher than that will appear white
 *	on the plot.
 */
#include <stdio.h>
#include <math.h>
#include "macro.h"
#include "Strings.h"

#define Nkeys 8		/* # of stops in key row */

float M[MATRIXSIZE][MATRIXSIZE] ;
int	m, n ;
float contrastfactor, treshold ;

FILE *fopen() ;

static findminmax(minv,maxv)
	float *minv, *maxv;
{
	int i, j ;
	float maxval,minval ;

	maxval = 0.0 ;
	minval = MAXINT ;
	for (i=0; i <= m ; i++)
		for (j=0; j <= n ; j++) {
			if (M[i][j] > maxval) maxval = M[i][j] ;
			if (M[i][j] < minval) minval = M[i][j] ;
		}
	*minv = minval ;
	*maxv = maxval ;
}


/*
 * linear in probabilities
 */
linearprobs()
{
	int i, j ;
	float minval, maxval ;
	int c ;

	findminmax(&minval,&maxval) ;

	for (i=0; i <= m ; i++) {
		for (j=0; j <= n ; j++) {
			if (M[i][j] - minval > 20) c = 255 ;
			else c = (int) ((1-exp2((double)minval-M[i][j]))*255+0.5) ;
		 	printf(" %3d",c) ;
		}
		putchar('\n') ;
	}
	printf("Probs\n") ;
	printf("%d ",Nkeys) ;
	for ( i = 0; i <= Nkeys; i++)
		printf("%3.2f ", 1.0-(float)i/Nkeys) ;
	putchar('\n') ;
}

/*
 *	Change matrix into grey scale value ... in linear scale
 */
linearscale()
{
	int i, j ;
	float minval, maxval, scale ;
	int c ;

	findminmax(&minval,&maxval) ;
	maxval -= minval ;
	treshold = fmin2(treshold,maxval) ;

	scale = 255.0/treshold ;

	for (i=0; i <= m ; i++) {
		for (j=0; j <= n ; j++) {
			if (M[i][j] - minval > treshold) c = 255 ;
			else c = (int) ((M[i][j]-minval)*scale+0.5) ;
		 	printf(" %3d",c) ;
		}
		putchar('\n') ;
	}
	printf("Mesg Length(bits)\n") ;
	printf("%d ",Nkeys) ;
	for ( i = 0; i <= Nkeys; i++) 
		printf("%d ",(int)(255*i/Nkeys/scale+0.5)) ;
	putchar('\n') ;
}

/*
 *	Change matrix into grey scale value ... in log scale
 */
logscale()
{
	int i, j ;
	float minval, maxval, scale ;
	int c ;

	findminmax(&minval,&maxval) ;
	maxval -= minval ;

	scale = (float)(255.0/log((double)maxval*contrastfactor+1)) ;
	for (i=0; i <= m ; i++) {
		for (j=0; j <= n ; j++) {
			c = (int) (log((double)(M[i][j]-minval)*contrastfactor+1)*scale+0.5) ;
		 	printf(" %3d",c) ;
		}
		putchar('\n') ;
	}
	printf("Mesg Length(bits)\n") ;
	printf("%d ",Nkeys) ;
	for ( i = 0; i <= Nkeys; i++) 
		printf("%d ",(int)((exp((double)255*i/Nkeys/scale)-1)/contrastfactor+0.5)) ;
	putchar('\n') ;
}

readln() { while (getchar() != '\n') ; }

/*
 *	read in the sequences and the density matrix
 */
ReadInput()
{
	int i, j ;
	char c ;

	scanf("%d %d",&m, &n) ;				/* length of strings */
	readln() ;
	if (m > MATRIXSIZE || n > MATRIXSIZE) {
		printf("ERROR: array too big\n") ;
		exit() ;
	}

	/* size of matrix, one larger than length of strings */
	printf("%d %d\n", m+1,n+1) ;

	for (i=1; i<=m ; ) 	/* the first sequence */
		if ( (c = getchar()) != '\n') { putchar(c) ; i++ ;}
	putchar('\n') ;
	readln() ;

	for (i=1; i<=n ; ) 	/* the second sequence */
		if ( (c = getchar()) != '\n') {putchar(c) ; i++ ;}
	putchar('\n') ;
	readln() ;
		
	for ( i = 0; i <= m; i++) { /* the density matrix */
		for ( j = 0; j <= n ; j++)
			scanf("%f",&M[i][j]) ;
		readln() ;
	}
}

main(argc, argv)
	int argc ;
	char *argv[] ;
{
	char 	c ;

	if (argc < 2 || argc > 3) {
		printf("usage: scale l(og) [contrast] | p | m [treshold] \n") ;
		exit(0) ;
	}
	switch (*argv[1]) {
	case 'l' :
		if (argc == 3) sscanf(argv[2],"%f",&contrastfactor) ;
		else contrastfactor = 1.0 ;
		ReadInput() ;
		logscale() ;
		break ;
	case 'p' :
		ReadInput() ;
		linearprobs() ;
		break ;
	case 'm' :
		if (argc == 3) sscanf(argv[2],"%f",&treshold) ;
		else treshold = 1000000.0 ;	/* should be large enough */
		ReadInput() ;
		linearscale() ;
		break ;
	default :
		printf("usage: scale l(og) [contrast] | p | m [treshold] \n") ;
		exit(0) ;
		break ;
	}

	/* the (optional) lines of caption */
	while ((c = getchar()) != '\n' && c != EOF) putchar(c) ;
	putchar('\n') ;
	while ((c = getchar()) != '\n' && c != EOF) putchar(c) ;
	putchar('\n') ;
}
